##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Objects.py,v 1.5.2.10 2009-07-24 13:35:53 ebke Exp $
##########################################################################
# NOTE: Make sure that _data and __dict__ of any GangaObject are only referenced
# here - this is necessary for write locking and lazy loading!
# IN THIS FILE:
# * Make sure every time _data or __dict__ is accessed that _data is not None. If yes, do:
#    obj._getReadAccess()
# * Make sure every write access is preceded with:
#    obj._getWriteAccess()
#   and followed by
#    obj._setDirty()

import threading
from contextlib import contextmanager
import functools

import Ganga.Utility.logging

from copy import deepcopy, copy
import inspect

import Ganga.GPIDev.Schema as Schema

from Ganga.Core.exceptions import GangaValueError, GangaException

from Ganga.Utility.Plugin import allPlugins

def _getName(obj):
    """ Return the name of an object based on what we prioritise"""
    returnable = getattr(obj, '_name', getattr(obj, '__name__', None))
    if returnable is None:
        returnable = getattr(getattr(obj, '__class__', None), '__name__', None)
    if returnable is None:
        returnable = str(obj)
    return returnable

logger = Ganga.Utility.logging.getLogger(modulename=1)

_imported_GangaList = None

do_not_copy = ['_index_cache', '_parent', '_registry', '_data', '_read_lock', '_write_lock', '_proxyObject']

def _getGangaList():
    global _imported_GangaList
    if _imported_GangaList is None:
        from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
        _imported_GangaList = GangaList
    return _imported_GangaList


def synchronised(f):
    """
    This decorator must be attached to a method on a ``Node`` subclass
    It uses the object's lock to make sure that the object is held for the duration of the decorated function
    """
    @functools.wraps(f)
    def decorated(self, *args, **kwargs):
        with self.const_lock:
            return f(self, *args, **kwargs)
    return decorated


class Node(object):
    _ref_list = ['_parent', '_registry', '_index_cache']

    def __init__(self, parent=None):
        self._data = {}
        self._parent = parent
        self._index_cache = {}
        self._read_lock = threading.RLock() # Don't read out of thread whilst we're making a change
        self._write_lock = threading.RLock() # Don't write from out of thread when modifying an object
        super(Node, self).__init__()
        #logger.info("Node __init__")

    def __getstate__(self):
        d = self.__dict__
        d['_data'] = d['_data'].copy()
        for r in self._ref_list:
            d[r] = None
        return d

    def __setstate__(self, this_dict):
        for key, val in this_dict['_data'].iteritems():
            if isinstance(val, Node) and key not in self._ref_list:
                val._setParent(self)

        for attr in self._ref_list:
            if not hasattr(self, attr):
                setattr(self, attr, None)

        for key, val in this_dict.iteritems():
            setattr(self, key, val)

    def __copy__(self, memo=None):
        cls = self.__class__
        obj = cls()
        # FIXME: this is different than for deepcopy... is this really correct?
        this_dict = self.__dict__.copy()
        global do_not_copy
        for elem in this_dict.keys():
            if elem not in do_not_copy:
                this_dict[elem] = copy(this_dict[elem])
            else:
                this_dict[elem] = None
        #obj.__setstate__(this_dict)
        obj._setParent(self._getParent())
        setattr(obj, '_index_cache', {})
        setattr(obj, '_registry', self._registry)
        return obj

    def __deepcopy__(self, memo=None):
        cls = self.__class__
        obj = cls()
        this_dict = self.__getstate__()
        global do_not_copy
        for elem in this_dict.keys():
            if elem not in do_not_copy:
                this_dict[elem] = deepcopy(this_dict[elem], memo)  # FIXED

        obj.__dict__ = this_dict
        if self._getParent() is not None:
            obj._setParent(self._getParent())
        setattr(obj, '_registry', self._registry)
        return obj

    def _getParent(self):
        return self._parent

    @synchronised  # This will lock the _current_ (soon to be _old_) root object
    def _setParent(self, parent):
        if parent is None:
            setattr(self, '_parent', parent)
        else:
            with parent.const_lock: # This will lock the _new_ root object
                setattr(self, '_parent', parent)
            # Finally the new and then old root objects will be unlocked

    @property
    @contextmanager
    def _internal_lock(self):
        """
        This is a context manager which acquires the internal read lock on the
        object's root object.
        """
        root = self._getRoot()
        root._read_lock.acquire()
        try:
            yield
        finally:
            root._read_lock.release()

    @property
    @contextmanager
    def const_lock(self):
        """
        This is a context manager which acquires the const write lock on the
        object's root object.

        This lock acquires exclusive access over an object tree to prevent it
        changing. Reading schema attributes on the object is still allowed
        but changing them is not. Only one thread can hold this lock at once.
        """
        root = self._getRoot()
        root._write_lock.acquire()
        try:
            yield
        finally:
            root._write_lock.release()

    # get the root of the object tree
    # if parent does not exist then the root is the 'self' object
    # cond is an optional function which may cut the search path: when it
    # returns True, then the parent is returned as root
    def _getRoot(self, cond=None):
        if self._getParent() is None:
            return self
        root = None
        obj = self
        cond_test = cond is not None
        while obj is not None:
            root = obj
            if cond_test:
                if cond(root):
                    break
            obj = obj._getParent()
        return root

    def _getdata(self, name):
        #logger.debug("Getting: %s" % name)
        if hasattr(self, name):
            return getattr(self, name)
        else:
            if name in self._data:
                return self._data[name]
            else:
                return None

    # accept a visitor pattern
    @synchronised
    def accept(self, visitor):

        if not hasattr(self, '_schema'):
            return
        elif self._schema is None:
            visitor.nodeBegin(self)
            visitor.nodeEnd(self)
            return

        visitor.nodeBegin(self)

        for (name, item) in self._schema.simpleItems():
            if item['visitable']:
                visitor.simpleAttribute(self, name, self._getdata(name), item['sequence'])

        for (name, item) in self._schema.sharedItems():
            if item['visitable']:
                visitor.sharedAttribute(self, name, self._getdata(name), item['sequence'])

        for (name, item) in self._schema.componentItems():
            if item['visitable']:
                visitor.componentAttribute(self, name, self._getdata(name), item['sequence'])

        visitor.nodeEnd(self)

    # clone self and return a properly initialized object
    def clone(self):
        new_obj = deepcopy(self)

        return new_obj

    # copy all the properties recursively from the srcobj
    # if schema of self and srcobj are not compatible raises a ValueError
    # ON FAILURE LEAVES SELF IN INCONSISTENT STATE
    def copyFrom(self, srcobj, _ignore_atts=None):

        if _ignore_atts is None:
            _ignore_atts = []
        _srcobj = srcobj
        # Check if this object is derived from the source object, then the copy
        # will not throw away information

        if not hasattr(_srcobj, '__class__') and not inspect.isclass(_srcobj.__class__):
            raise GangaValueError("Can't copyFrom a non-class object: %s isclass: %s" % (str(_srcobj), str(inspect.isclass(_srcobj))))

        if not isinstance(self, _srcobj.__class__) and not isinstance(_srcobj, self.__class__):
            raise GangaValueError("copyFrom: Cannot copy from %s to %s!" % (_getName(_srcobj), _getName(self)))

        if not hasattr(self, '_schema'):
            logger.debug("No Schema found for myself")
            return

        if self._schema is None and _srcobj._schema is None:
            logger.debug("Schema object for one of these classes is None!")
            return

        if _srcobj._schema is None:
            self._schema = None
            return

        self._actually_copyFrom(_srcobj, _ignore_atts)

        ## Fix some objects losing parent knowledge
        src_dict = srcobj.__dict__
        for key, val in src_dict.iteritems():
            this_attr = getattr(srcobj, key)
            if isinstance(this_attr, Node) and key not in Node._ref_list:
                #logger.debug("k: %s  Parent: %s" % (str(key), (srcobj)))
                this_attr._setParent(srcobj)

    def _actually_copyFrom(self, _srcobj, _ignore_atts):

        for name, item in self._schema.allItems():
            if name in _ignore_atts:
                continue

            #logger.debug("Copying: %s : %s" % (str(name), str(item)))
            if name == 'application' and hasattr(_srcobj.application, 'is_prepared'):
                _app = _srcobj.application
                if _app.is_prepared not in [None, True]:
                    _app.incrementShareCounter(_app.is_prepared.name)

            if not self._schema.hasAttribute(name):
                #raise ValueError('copyFrom: incompatible schema: source=%s destination=%s'%(_getName(_srcobj), _getName(self)))
                if not hasattr(self, name):
                    setattr(self, name, self._schema.getDefaultValue(name))
                this_attr = getattr(self, name)
                if isinstance(this_attr, Node) and name not in Node._ref_list:
                    this_attr._setParent(self)
            elif not item['copyable']: ## Default of '1' instead of True...
                if not hasattr(self, name):
                    setattr(self, name, self._schema.getDefaultValue(name))
                this_attr = getattr(self, name)
                if isinstance(this_attr, Node) and name not in Node._ref_list:
                    this_attr._setParent(self)
            else:
                copy_obj = deepcopy(getattr(_srcobj, name))
                setattr(self, name, copy_obj)

    def printTree(self, f=None, sel=''):
        from Ganga.GPIDev.Base.VPrinter import VPrinter
        self.accept(VPrinter(f, sel))

    #printPrepTree is only ever run on applications, from within IPrepareApp.py
    #if you (manually) try to run printPrepTree on anything other than an application, it will not work as expected
    #see the relevant code in VPrinter to understand why
    def printPrepTree(self, f=None, sel='preparable' ):
        ## After fixing some bugs we are left with incompatible job hashes. This should be addressd before removing
        ## This particular class!
        from Ganga.GPIDev.Base.VPrinterOld import VPrinterOld
        self.accept(VPrinterOld(f, sel))

    def printSummaryTree(self, level=0, verbosity_level=0, whitespace_marker='', out=None, selection='', interactive=False):
        """If this method is overridden, the following should be noted:

        level: the hierachy level we are currently at in the object tree.
        verbosity_level: How verbose the print should be. Currently this is always 0.
        whitespace_marker: If printing on multiple lines, this allows the default indentation to be replicated.
                           The first line should never use this, as the substitution is 'name = %s' % printSummaryTree()
        out: An output stream to print to. The last line of output should be printed without a newline.'
        selection: See VPrinter for an explaintion of this.
        """
        from Ganga.GPIDev.Base.VPrinter import VSummaryPrinter
        self.accept(VSummaryPrinter(level, verbosity_level, whitespace_marker, out, selection, interactive))

    def __eq__(self, _node):

        node = _node

        if self is node:
            return 1
        if not node:  # or not self._schema.isEqual(node._schema):
            return 0

        if not isinstance(node, type(self)):
            return 0

        # Compare the schemas against each other
        if (hasattr(self, '_schema') and self._schema is None) and (hasattr(node, '_schema') and node._schema is None):
            return 1  # If they're both `None`
        elif (hasattr(self, '_schema') and self._schema is None) or (hasattr(node, '_schema') and node._schema is None):
            return 0  # If just one of them is `None`
        elif not self._schema.isEqual(node._schema):
            return 0  # Both have _schema but do not match

        # Check each schema item in turn and check for equality
        for (name, item) in self._schema.allItems():
            if item['comparable'] == True:
                #logger.info("testing: %s::%s" % (str(_getName(self)), str(name)))
                if getattr(self, name) != getattr(node, name):
                    #logger.info( "diff: %s::%s" % (str(_getName(self)), str(name)))
                    return 0

        return 1

    def __ne__(self, node):
        return not self.__eq__(node)

    def getNodeData(self):
        return self._data

    def setNodeData(self, new_data):
        self._data = new_data
        for k, v in self._data.iteritems():
            if isinstance(v, Node):
                v._setParent(self)

    def getNodeAttribute(self, attrib_name):
        return self.getNodeData()[attrib_name]

    def setNodeAttribute(self, attrib_name, attrib_value):
        self.getNodeData()[attrib_name] = attrib_value
        ## ALL of the functions are in the NODE class not GANGAOBJECT
        if isinstance(attrib_value, Node):
            self.getNodeData()[attrib_name]._setParent(self)

    def removeNodeAttribute(self, attrib_name):
        if attrib_name in self._data.keys():
            del self._data[attrib_name]

    def removeNodeIndexCacheAttribute(self, attrib_name):
        if self._index_cache and attrib_name in self._index_cache.keys():
            del self._index_cache[attrib_name]

    def setNodeIndexCache(self, new_index_cache):
        setattr(self, '_index_cache', new_index_cache)

    def getNodeIndexCache(self):
        return self._index_cache

##########################################################################


def synchronised_get_descriptor(get_function):
    """
    This decorator should only be used on ``__get__`` method of the ``Descriptor``.
    """
    @functools.wraps(get_function)
    def decorated(self, obj, type_or_value):
        if obj is None:
            return get_function(self, obj, type_or_value)

        with obj._internal_lock:
            return get_function(self, obj, type_or_value)

    return decorated

def synchronised_set_descriptor(set_function):
    """
    This decorator should only be used on ``__set__`` method of the ``Descriptor``.
    """
    def decorated(self, obj, type_or_value):
        if obj is None:
            return set_function(self, obj, type_or_value)

        with obj.const_lock:
            with obj._internal_lock:
                return set_function(self, obj, type_or_value)
    return decorated


class Descriptor(object):

    def __init__(self, name, item):
        self._name = name
        self._item = item
        self._getter_name = None
        self._checkset_name = None
        self._filter_name = None

        self._getter_name = item['getter']
        self._checkset_name = item['checkset']
        self._filter_name = item['filter']

    @staticmethod
    def _bind_method(obj, name):
        if name is None:
            return None
        return getattr(obj, name)

    def _check_getter(self):
        if self._getter_name:
            raise AttributeError('cannot modify or delete "%s" property (declared as "getter")' % _getName(self))

    @synchronised_get_descriptor
    def __get__(self, obj, cls):
        name = _getName(self)

        # If obj is None then the getter was called on the class so return the Item
        if obj is None:
            return cls._schema[name]

        if self._getter_name:
            return self._bind_method(obj, self._getter_name)()

        # First we want to try to get the information without prompting a load from disk

        # ._data takes priority ALWAYS over ._index_cache
        # This access should not cause the object to be loaded
        obj_data = obj.getNodeData()
        if name in obj_data:
            return obj_data[name]

        # Then try to get it from the index cache
        obj_index = obj.getNodeIndexCache()
        if name in obj_index:
            return obj_index[name]

        # Since we couldn't find the information in the cache, we will need to fully load the object

        # Guarantee that the object is now loaded from disk
        obj._getReadAccess()

        # First try to load the object from the attributes on disk
        if name in obj.getNodeData():
            return obj.getNodeAttribute(name)

        # Finally, get the default value from the schema
        if obj._schema.hasItem(name):
            return obj._schema.getDefaultValue(name)

        raise AttributeError('Could not find attribute {0} in {1}'.format(name, obj))

    def __cloneVal(self, v, obj):

        item = obj._schema[_getName(self)]

        if v is None:
            if item.hasProperty('category'):
                assertion = item['optional'] and (item['category'] != 'internal')
            else:
                assertion = item['optional']
            #assert(assertion)
            if assertion is False:
                logger.warning("Item: '%s'. of class type: '%s'. Has a Default value of 'None' but is NOT optional!!!" % (_getName(self), type(obj)))
                logger.warning("Please contact the developers and make sure this is updated!")
            return None
        elif isinstance(v, str):
            return str(v)
        elif isinstance(v, int):
            return int(v)
        elif isinstance(v, dict):
            new_dict = {}
            for key, item in new_dict.iteritems():
                new_dict[key] = self.__cloneVal(v, obj)
            return new_dict
        else:
            if not isinstance(v, Node) and isinstance(v, (list, tuple)):
                try:
                    GangaList = _getGangaList()
                    new_v = GangaList()
                except ImportError:
                    new_v = []
                for elem in v:
                    new_v.append(self.__cloneVal(elem, obj))
                #return new_v
            elif not isinstance(v, Node):
                if inspect.isclass(v):
                    new_v = v()
                else:
                    new_v = v
                if not isinstance(new_v, Node):
                    logger.error("v: %s" % str(v))
                    raise GangaException("Error: found Object: %s of type: %s expected an object inheriting from Node!" % (str(v), str(type(v))))
                else:
                    new_v = self.__copyNodeObject(new_v, obj)
            else:
                new_v = self.__copyNodeObject(v, obj)

            return new_v

    def __copyNodeObject(self, v, obj):
        """This deals with the actual deepcopy of an object which has inherited from Node class"""

        item = obj._schema[_getName(self)]
        GangaList = _getGangaList()
        if isinstance(v, GangaList):
            categories = v.getCategory()
            len_cat = len(categories)
            if (len_cat > 1) or ((len_cat == 1) and (categories[0] != item['category'])) and item['category'] != 'internal':
                # we pass on empty lists, as the catagory is yet to be defined
                from Ganga.GPIDev.Base.Proxy import GangaAttributeError
                raise GangaAttributeError('%s: attempt to assign a list containing incompatible objects %s to the property in category "%s"' % (_getName(self), v, item['category']))
        else:
            if v._category not in [item['category'], 'internal'] and item['category'] != 'internal':
                from Ganga.GPIDev.Base.Proxy import GangaAttributeError
                raise GangaAttributeError('%s: attempt to assign an incompatible object %s to the property in category "%s found cat: %s"' % (_getName(self), v, item['category'], v._category))


        v_copy = deepcopy(v)

        #logger.info("Cloned Object Parent: %s" % v_copy._getParent())
        #logger.info("Original: %s" % v_copy._getParent())

        return v_copy

    @synchronised_set_descriptor
    def __set__(self, _obj, _val):
        ## self: attribute being changed or Ganga.GPIDev.Base.Objects.Descriptor in which case _getName(self) gives the name of the attribute being changed
        ## _obj: parent class which 'owns' the attribute
        ## _val: value of the attribute which we're about to set

        obj_reg = None
        obj_prevState = None
        obj = _obj
        if isinstance(obj, GangaObject):
            obj_reg = obj._getRegistry()
            if obj_reg is not None and hasattr(obj_reg, 'isAutoFlushEnabled'):
                obj_prevState = obj_reg.isAutoFlushEnabled()
                if obj_prevState is True and hasattr(obj_reg, 'turnOffAutoFlushing'):
                    obj_reg.turnOffAutoFlushing()

        val_reg = None
        val_prevState = None
        val = _val
        if isinstance(val, GangaObject):
            val_reg = val._getRegistry()
            if val_reg is not None and hasattr(val_reg, 'isAutoFlushEnabled'):
                val_prevState = val_reg.isAutoFlushEnabled()
                if val_prevState is True and hasattr(val_reg, 'turnOffAutoFlushing'):
                    val_reg.turnOffAutoFlushing()

        if type(_val) is str:
            from Ganga.GPIDev.Base.Proxy import stripProxy, runtimeEvalString
            new_val = stripProxy(runtimeEvalString(_obj, _getName(self), _val))
        else:
            new_val = _val

        self.__atomic_set__(_obj, new_val)

        if isinstance(new_val, Node):
            val._setDirty()

        if val_reg is not None:
            if val_prevState is True and hasattr(val_reg, 'turnOnAutoFlushing'):
                val_reg.turnOnAutoFlushing()

        if obj_reg is not None:
            if obj_prevState is True and hasattr(obj_reg, 'turnOnAutoFlushing'):
                obj_reg.turnOnAutoFlushing()

    def __atomic_set__(self, _obj, _val):
        ## self: attribute being changed or Ganga.GPIDev.Base.Objects.Descriptor in which case _getName(self) gives the name of the attribute being changed
        ## _obj: parent class which 'owns' the attribute
        ## _val: value of the attribute which we're about to set

        #if hasattr(_obj, _getName(self)):
        #    if not isinstance(getattr(_obj, _getName(self)), GangaObject):
        #        if type( getattr(_obj, _getName(self)) ) == type(_val):
        #            object.__setattr__(_obj, _getName(self), deepcopy(_val))
        #            return
#
#        if not isinstance(_obj, GangaObject) and type(_obj) == type(_val):
#            _obj = deepcopy(_val)
#            return

        obj = _obj
        temp_val = _val

        from Ganga.GPIDev.Lib.GangaList.GangaList import makeGangaList

        if hasattr(obj, '_checkset_name'):
            checkSet = self._bind_method(obj, self._checkset_name)
            if checkSet is not None:
                checkSet(temp_val)
        if hasattr(obj, '_filter_name'):
            this_filter = self._bind_method(obj, self._filter_name)
            if this_filter:
                val = this_filter(temp_val)
            else:
                val = temp_val
        else:
            val = temp_val

        # LOCKING
        obj._getWriteAccess()

        #self._check_getter()

        item = obj._schema[_getName(self)]

        def cloneVal(v):
            GangaList = _getGangaList()
            if isinstance(v, (list, tuple, GangaList)):
                new_v = GangaList()
                for elem in v:
                    new_v.append(self.__cloneVal(elem, obj))
                return new_v
            else:
                return self.__cloneVal(v, obj)

        ## If the item has been defined as a sequence great, let's continue!
        if item['sequence']:
            _preparable = True if item['preparable'] else False
            if len(val) == 0:
                GangaList = _getGangaList()
                new_val = GangaList()
            else:
                if isinstance(item, Schema.ComponentItem):
                    new_val = makeGangaList(val, cloneVal, parent=obj, preparable=_preparable)
                else:
                    new_val = makeGangaList(val, parent=obj, preparable=_preparable)
        else:
            ## Else we need to work out what we've got.
            if isinstance(item, Schema.ComponentItem):
                GangaList = _getGangaList()
                if isinstance(val, (list, tuple, GangaList)):
                    ## Can't have a GangaList inside a GangaList easily so lets not
                    if isinstance(_obj, GangaList):
                        newListObj = []
                    else:
                        newListObj = GangaList()

                    Descriptor.__createNewList(newListObj, val, cloneVal)
                    #for elem in val:
                    #    newListObj.append(cloneVal(elem))
                    new_val = newListObj
                else:
                    new_val = cloneVal(val)
            else:
                new_val = val
                pass
            #val = deepcopy(val)

        if isinstance(new_val, Node):
            new_val._setParent(obj)

        obj.setNodeAttribute(_getName(self), new_val)

        obj._setDirty()

    def __delete__(self, obj):
        obj.removeNodeAttribute(_getName(self))

    @staticmethod
    def __createNewList(final_list, input_elements, action=None):

        def addToList(_input_elements, _final_list, action=None):
            if action is not None:
                for element in _input_elements:
                    _final_list.append(action(element))
            else:
                for element in _input_elements:
                    _final_list.append(element)
            return

        ## This makes it stick to 1 thread, useful for debugging problems
        addToList(input_elements, final_list, action)
        return


class ObjectMetaclass(type):
    _descriptor = Descriptor

    def __init__(cls, name, bases, this_dict):

        super(ObjectMetaclass, cls).__init__(name, bases, this_dict)

        # all Ganga classes must have (even empty) schema
        if cls._schema is None:
            cls._schema = Schema.Schema(None, None)

        this_schema = cls._schema

        # Add all class members of type `Schema.Item` to the _schema object
        # TODO: We _could_ add base class's Items here by going through `bases` as well.
        # We can't just yet because at this point the base class' Item has been overwritten with a Descriptor
        for member_name, member in this_dict.items():
            if isinstance(member, Schema.Item):
                this_schema.datadict[member_name] = member

        # sanity checks for schema...
        if '_schema' not in this_dict.keys():
            s = "Class %s must _schema (it cannot be silently inherited)" % (name,)
            logger.error(s)
            raise ValueError(s)

        # If a class has not specified a '_name' then default to using the class '__name__'
        if not cls.__dict__.get('_name'):
            cls._name = name

        if this_schema._pluginclass is not None:
            logger.warning('Possible schema clash in class %s between %s and %s', name, _getName(cls), _getName(this_schema._pluginclass))

        # export visible properties... do not export hidden properties
        for attr, item in this_schema.allItems():
            setattr(cls, attr, cls._descriptor(attr, item))

        # additional check of type
        # bugfix #40220: Ensure that default values satisfy the declared types
        # in the schema
        for attr, item in this_schema.simpleItems():
            if not item['getter']:
                item._check_type(item['defvalue'], '.'.join([name, attr]), enableGangaList=False)

        # create reference in schema to the pluginclass
        this_schema._pluginclass = cls

        # if we've not even declared this we don't want to use it!
        if not cls._declared_property('hidden') or cls._declared_property('enable_plugin'):
            allPlugins.add(cls, cls._category, _getName(cls))

        # create a configuration unit for default values of object properties
        if not cls._declared_property('hidden') or cls._declared_property('enable_config'):
            this_schema.createDefaultConfig()


class GangaObject(Node):
    __metaclass__ = ObjectMetaclass
    _schema = None  # obligatory, specified in the derived classes
    _category = None  # obligatory, specified in the derived classes
    _registry = None  # automatically set for Root objects
    _exportmethods = []  # optional, specified in the derived classes

    # by default classes are not hidden, config generation and plugin
    # registration is enabled
    # optional, specify in the class if you do not want to export it publicly
    # in GPI,
    _hidden = 1
    # the class will not be registered as a plugin unless _enable_plugin is defined
    # the test if the class is hidden is performed by x._declared_property('hidden')
    # which makes sure that _hidden must be *explicitly* declared, not
    # inherited

    # additional properties that may be set in derived classes which were declared as _hidden:
    #   _enable_plugin = 1 -> allow registration of _hidden classes in the allPlugins dictionary
    # _enable_config = 1 -> allow generation of [default_X] configuration
    # section with schema properties

    # must be fully initialized
    def __init__(self):
        super(GangaObject, self).__init__(None)

        # IMPORTANT: if you add instance attributes like in the line below
        # make sure to update the __getstate__ method as well
        # dirty flag is true if the object has been modified locally and its
        # contents is out-of-sync with its repository
        self._dirty = False

        #Node.__init__(self, None)

        if self._schema is not None and hasattr(self._schema, 'allItems'):
            for attr, item in self._schema.allItems():
                ## If an object is hidden behind a getter method we can't assign a parent or defvalue so don't bother - rcurrie
                if item.getProperties()['getter'] is None:
                    defVal = self._schema.getDefaultValue(attr)
                    self.setNodeAttribute(attr, defVal)

        # Overwrite default values with any config values specified
        # self.setPropertiesFromConfig()

    def __construct__(self, args):
        # type: (Sequence) -> None
        """
        This acts like a secondary constructor for proxy objects.
        Any positional (non-keyword) arguments are passed to this function to construct the object.

        This default implementation performs a copy if there was only one item in the list
        and raises an exception if there is more than one.

        Args:
            args: a list of objects

        Raises:
            TypeMismatchError: if there is more than one item in the list
        """
        # FIXME: This should probably be move to Proxy.py

        if len(args) == 0:
            return
        elif len(args) == 1:
            if not isinstance(args[0], type(self)):
                logger.warning("Performing a copyFrom from: %s to: %s" % (type(args[0]), type(self)))
            self.copyFrom(args[0])
        else:
            from Ganga.GPIDev.Base.Proxy import TypeMismatchError
            raise TypeMismatchError("Constructor expected one or zero non-keyword arguments, got %i" % len(args))

    def __getstate__(self):
        # IMPORTANT: keep this in sync with the __init__
        #self._getReadAccess()
        this_dict = super(GangaObject, self).__getstate__()
        #this_dict['_dirty'] = False
        return this_dict

    def __setstate__(self, this_dict):
        #self._getWriteAccess()
        super(GangaObject, self).__setstate__(this_dict)
        #if '_parent' in this_dict:
        #    self._setParent(this_dict['_parent'])
        #self._setParent(None)
        self._dirty = False

    @staticmethod
    def __incrementShareRef(obj, attr_name):
        shared_dir = getattr(obj, attr_name)

        if hasattr(shared_dir, 'name'):

            from Ganga.Core.GangaRepository import getRegistry
            from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory
            shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())

            logger.debug("Increasing shareref")
            shareref.increase(shared_dir.name)

    # on the deepcopy reset all non-copyable properties as defined in the
    # schema
    def __deepcopy__(self, memo=None):
        true_parent = self._getParent()
        ## This triggers a read of the job from disk
        self._getReadAccess()
        classname = _getName(self)
        category = self._category
        cls = self.__class__#allPlugins.find(category, classname)

        self_copy = cls()

        global do_not_copy
        if self._schema is not None:
            for name, item in self._schema.allItems():
                if not item['copyable'] or name in do_not_copy:
                    setattr(self_copy, name, self._schema.getDefaultValue(name))
                else:
                    if hasattr(self, name):
                        setattr(self_copy, name, deepcopy(getattr(self, name)))
                    else:
                        setattr(self_copy, name, self._schema.getDefaultValue(name))

                this_attr = getattr(self_copy, name)
                if isinstance(this_attr, Node):
                    this_attr._setParent(self_copy)

                if item.isA(Schema.SharedItem):
                    self.__incrementShareRef(self_copy, name)

        for k, v in self.__dict__.iteritems():
            if k not in do_not_copy:
                try:
                    self_copy.__dict__[k] = deepcopy(v)
                except:
                    self_copy.__dict__[k] = v

        if true_parent is not None:
            self._setParent(true_parent)
            self_copy._setParent(true_parent)
        return self_copy

    def _getIOTimeOut(self):
        from Ganga.Utility.Config.Config import getConfig, ConfigError
        try:
            _timeOut = getConfig('Configuration')['DiskIOTimeout']
        except ConfigError, err:
            _timeOut = 5. # 5sec hardcoded default
        return _timeOut

    def _getWriteAccess(self):
        """ tries to get write access to the object.
        Raise LockingError (or so) on fail """
        root = self._getRoot()
        reg = root._getRegistry()
        if reg is not None:
            _haveLocked = False
            _counter = 1
            _sleep_size = 1
            _timeOut = self._getIOTimeOut()
            while not _haveLocked:
                err = None
                from Ganga.Core.GangaRepository.Registry import RegistryLockError, RegistryAccessError
                reg._write_access(root)
                try:
                    reg._write_access(root)
                    _haveLocked = True
                except (RegistryLockError, RegistryAccessError) as x:
                    #import traceback
                    #traceback.print_stack()
                    from time import sleep
                    sleep(_sleep_size)  # Sleep 2 sec between tests
                    logger.info("Waiting on Write access to registry: %s" % reg.name)
                    logger.debug("err: %s" % str(x))
                    err = x
                _counter = _counter + 1
                # Sleep 2 sec longer than the time taken to bail out
                if _counter * _sleep_size >= _timeOut + 2:
                    logger.error("Failed to get access to registry: %s. Reason: %s" % (reg.name, str(err)))
                    if err is not None:
                        raise err

    def _releaseWriteAccess(self):
        """ releases write access to the object.
        Raise LockingError (or so) on fail
        Please use only if the object is expected to be used by other sessions"""
        root = self._getRoot()
        reg = root._getRegistry()
        if reg is not None:
            logger.debug("Releasing: %s" % (reg.name))
            reg._release_lock(root)

    def _getReadAccess(self):
        """ makes sure the objects _data is there and the object itself has a recent state.
        Raise RepositoryError"""
        root = self._getRoot()
        reg = root._getRegistry()
        if reg is not None:
            reg._read_access(root, self)

    # define when the object is read-only (for example a job is read-only in
    # the states other than new)
    def _readonly(self):
        r = self._getRoot()
        # is object a root for itself? check needed otherwise infinite
        # recursion
        if r is None or r is self:
            return 0
        else:
            return r._readonly()

    # set the registry for this object (assumes this object is a root object)
    def _setRegistry(self, registry):
        assert self._getParent() is None
        self._registry = registry

    # get the registry for the object by getting the registry associated with
    # the root object (if any)
    def _getRegistry(self):
        r = self._getRoot()
        return r._registry

    def _getRegistryID(self):
        try:
            return self._registry.find(self)
        except AttributeError, err:
            logger.debug("_getRegistryID Exception: %s" % str(err))
            return None

    # mark object as "dirty" and inform the registry about it
    # the registry is always associated with the root object
    def _setDirty(self, dummy=1):
        self._dirty = True
        parent = self._getParent()
        if parent is not None:
            parent._setDirty()

    def _setFlushed(self):
        self._dirty = False

    # post __init__ hook automatically called by GPI Proxy __init__
    def _auto__init__(self):
        pass

    # return True if _name attribute was explicitly defined in the class
    # this means that implicit (inherited) _name attribute has no effect in the derived class
    # example: cls._declared_property('hidden') => True if there is class
    # attribute _hidden explicitly declared
    @classmethod
    def _declared_property(cls, name):
        return '_' + name in cls.__dict__

    # get the job object associated with self or raise an assertion error
    # the FIRST PARENT Job is returned...
    # this method is for convenience and may well be moved to some subclass
    def getJobObject(self):
        from Ganga.GPIDev.Lib.Job import Job
        r = self._getRoot(cond=lambda o: isinstance(o, Job))
        if not isinstance(r, Job):
            raise AssertionError('no job associated with object ' + repr(self))
        return r

    # Customization of the GPI attribute assignment: Attribute Filters
    #
    # Example of usage:
    # if some properties depend on the value of other properties in a complex way such as:
    # changing platform of Gaudi should change the version if it is not supported... etc.
    #
    # Semantics:
    #  gpi_proxy.x = v        --> stripProxy(gpi_proxy)._attribute_filter__set__('x',v)
    #  gpi_proxy.y = [v1,v2]  --> stripProxy(gpi_proxy)._attribute_filter__set__('x',[v1,v2])
    #
    #  is used for *all* kinds of attributes (component and simple)
    #
    # Attribute Filters are used mainly for side effects such as modification of the state of the self object
    # (as described above).
    #
    # Attribute Filter may also perform an additional conversion of the value which is being assigned.
    #
    # Returns the attribute value (converted or original)
    #
    def _attribute_filter__set__(self, name, v):
        if hasattr(v, '_on_attribute__set__'):
            return v._on_attribute__set__(self, name)
        return v


# define the default component object filter:
# obj.x = "Y"   <=> obj.x = Y()


def string_type_shortcut_filter(val, item):
    if isinstance(val, type('')):
        if item is None:
            raise ValueError('cannot apply default string conversion, probably you are trying to use it in the constructor')
        from Ganga.Utility.Plugin import allPlugins, PluginManagerError
        try:
            obj = allPlugins.find(item['category'], val)()
            obj._auto__init__()
            return obj
        except PluginManagerError as err:
            logger.debug("string_type_shortcut_filter Exception: %s" % str(err))
            raise ValueError(err)
    return None

# FIXME: change into classmethod (do they inherit?) and then change stripComponentObject to use class instead of
# FIXME: object (object model clearly fails with sequence of Files)
# FIXME: test: ../bin/ganga -c local_lhcb.ini run.py
# TestNativeSpecific.testFileSequence


from .Filters import allComponentFilters
allComponentFilters.setDefault(string_type_shortcut_filter)

#
#
# $Log: not supported by cvs2svn $
# Revision 1.5.2.9  2009/07/14 14:44:17  ebke
# * several bugfixes
# * changed indexing for XML/Pickle
# * introduce index update minimal time of 20 seconds (reduces lag for typing 'jobs')
# * subjob splitting and individual flushing for XML/Pickle
#
# Revision 1.5.2.8  2009/07/13 22:10:53  ebke
# Update for the new GangaRepository:
# * Moved dict interface from Repository to Registry
# * Clearly specified Exceptions to be raised by Repository
# * proper exception handling in Registry
# * moved _writable to _getWriteAccess, introduce _getReadAccess
# * clarified locking, logic in Registry, less in Repository
# * index reading support in XML (no writing, though..)
# * general index reading on registry.keys()
#
# Revision 1.5.2.7  2009/07/10 12:14:10  ebke
# Fixed wrong sequence in __set__: only dirty _after_ writing!
#
# Revision 1.5.2.6  2009/07/10 11:33:06  ebke
# Preparations and fixes for lazy loading
#
# Revision 1.5.2.5  2009/07/08 15:27:50  ebke
# Removed load speed bottleneck for pickle - reduced __setstate__ time by factor 3.
#
# Revision 1.5.2.4  2009/07/08 12:51:52  ebke
# Fixes some bugs introduced in the latest version
#
# Revision 1.5.2.3  2009/07/08 12:36:54  ebke
# Simplified _writable
#
# Revision 1.5.2.2  2009/07/08 11:18:21  ebke
# Initial commit of all - mostly small - modifications due to the new GangaRepository.
# No interface visible to the user is changed
#
# Revision 1.5.2.1  2009/06/04 12:00:37  moscicki
# *** empty log message ***
#
# Revision 1.5  2009/05/20 13:40:22  moscicki
# added filter property for GangaObjects
#
# added Utility.Config.expandgangasystemvars() filter which expands @{VAR} in strings, where VAR is a configuration option defined in the System section
# Usage example: specify @{GANGA_PYTHONPATH} in the configuration file to make pathnames relative to the location of Ganga release; specify @{GANGA_VERSION} to expand to current Ganga version. etc.
#
# modified credentials package (ICommandSet) to use the expandgangasystemvars() filter.
#
# Revision 1.4  2009/04/27 09:22:56  moscicki
# fix #29745: use __mro__ rather than first-generation of base classes
#
# Revision 1.3  2009/02/24 14:57:56  moscicki
# set parent correctly for GangaList items (in __setstate__)
#
# Revision 1.2  2008/09/09 14:37:16  moscicki
# bugfix #40220: Ensure that default values satisfy the declared types in the schema
#
# factored out type checking into schema module, fixed a number of wrongly declared schema items in the core
#
# Revision 1.1  2008/07/17 16:40:52  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.27.4.10  2008/03/31 15:30:26  kubam
# More flexible internal logic of hiding the plugin classes derived from GangaObject:
#
# # by default classes are not hidden, config generation and plugin registration is enabled
#
# _hidden       = 1  # optional, specify in the class if you do not want to export it publicly in GPI,
#                    # the class will not be registered as a plugin unless _enable_plugin is defined
#                    # the test if the class is hidden is performed by x._declared_property('hidden')
#
# # additional properties that may be set in derived classes which were declared as _hidden:
# #   _enable_plugin = 1 -> allow registration of _hidden classes in the allPlugins dictionary
# #   _enable_config = 1 -> allow generation of [default_X] configuration section with schema properties
#
# This fixes: bug #34470: config [defaults_GridProxy] missing (issue of _hidden property of GangaObject)
#
# Revision 1.27.4.9  2008/02/28 15:44:31  moscicki
# fixed set parent problem for GangaList (and removed Will's hack which was already commented out)
#
# Revision 1.27.4.8  2008/02/12 09:25:24  amuraru
# removed Will's hack to set the parent, it causes side effects in subjobs accessor(to be checked)
#
# Revision 1.27.4.7  2008/02/06 17:02:00  moscicki
# small fix
#
# Revision 1.27.4.6  2008/02/06 13:00:21  wreece
# Adds a bit of a HACK as a tempory measure. The parent of a GangaList is being lost somewhere (I suspect due to a copy). I've added it back manually in __get__.
#
# Revision 1.27.4.5  2008/02/06 09:28:48  wreece
# First pass at a cleanup of the gangalist stuff. I've made changes so the diffs with the 4.4 series are more transparent. Currently still test failures.
#
# Revision 1.27.4.4  2007/12/18 09:05:04  moscicki
# integrated typesystem from Alvin and made more uniform error reporting
#
# Revision 1.27.4.3  2007/11/14 13:03:54  wreece
# Changes to make shortcuts work correctly with gangalists. all but one tests should now pass.
#
# Revision 1.27.4.2  2007/11/07 15:10:02  moscicki
# merged in pretty print and GangaList support from ganga-5-dev-branch-4-4-1-will-print branch
#
#
# Revision 1.27.4.1  2007/10/30 15:25:53  moscicki
# fixed #29745: Inherited functions of GangaObjects cannot be exported via _exportmethods
#
# Revision 1.27.8.3  2007/10/30 21:22:02  wreece
# Numerous small fixes and work arounds. And a new test.
#
# Revision 1.27.8.2  2007/10/30 14:30:23  wreece
# Non-working update. Adds in Kuba's exported methods dodge. It is now possible to define a _export_ version of a method for external use and a undecorated method for internal use.
#
# Revision 1.27.8.1  2007/10/30 12:12:08  wreece
# First version of the new print_summary functionality. Lots of changes, but some known limitations. Will address in next version.
#
# Revision 1.27  2007/07/27 13:52:00  moscicki
# merger updates from Will (from GangaMergers-1-0 branch)
#
# Revision 1.26.12.1  2007/05/14 13:32:11  wreece
# Adds the merger related code on a seperate branch. All tests currently
# run successfully.
#
# Revision 1.26  2006/07/27 20:09:54  moscicki
# _getJobObject() renamed -> getJobObject() and finding FIRST PARENT JOB
# _getRoot() : cond optional argument (to cut the search path e.g. for getJobObject())
#
# getDefaultValue() moved to Schema object
#
# "checkset" metaproperty implemented in schema (to check the direct updates to job.status)
# modified the "getter" metaproperty implementation
# create automatically the configuration units with default property values
#
# Revision 1.25  2006/02/10 14:18:44  moscicki
# removed obsoleted eval for default properties from config file
#
# Revision 1.24  2006/01/09 16:36:55  moscicki
#  - support for defining default properties in the configuration
#    config file example:
#
#      [LSF_Properties]
#      queue = myQueue
#
#      [LCG_Properties]
#      requirements.CE = myCE
#
#      [Job_Properties]
#      application = DaVinci
#      application.version = v99r2
#
# Revision 1.23  2005/12/02 15:27:13  moscicki
# support for "visitable" metaproperty
# support for hidden GPI classes
# "hidden" properties not visible in the proxies
# support for 'getter' type of property (a binding) for job.master
#
# Revision 1.22  2005/11/14 10:29:55  moscicki
# adapted to changed interface of allPlugins.add() method
#
# Revision 1.21  2005/11/01 16:43:51  moscicki
# added isStringLike condition to avoid unnecesary looping over strings
#
# Revision 1.20  2005/11/01 14:04:55  moscicki
# added missing implementation of GangaObject.__setstate__
# fixed the implementation of Node.__setstate__ to correctly set parent for objects which are contained in iterable simple items (lists, dictionaries) etc. (contributed by AS)
#
# Revision 1.19  2005/11/01 11:09:48  moscicki
# changes to support import/export (KH):
#   - The printTree method accepts an additional argument, which
#     it passes on to VPrinter as the value for "selection".
#
# Revision 1.18  2005/08/26 09:53:24  moscicki
# copy __doc__ for exported method
# _getJobObject() helper method in GangaObject
#
# Revision 1.17  2005/08/24 15:41:19  moscicki
# automatically generated help for properties, disabled the SchemaHelper and few other improvements to the help system
#
# Revision 1.16  2005/08/23 17:15:06  moscicki
# *** empty log message ***
#
#
#



#!/usr/bin/env python

import os
from dq2.clientapi.DQ2 import DQ2
from dq2.info import TiersOfATLAS
from dq2.info.TiersOfATLAS import _refreshToACache, ToACache

_refreshToACache()

dq2localid = os.environ['DQ2_LOCAL_SITE_ID']
dq2localid_alternatename = TiersOfATLAS.getSiteProperty(dq2localid,'alternateName')
if not dq2localid_alternatename:
    try:
        dq2localid_alternatename = dq2localid.split("_")[0]
    except:
        dq2localid_alternatename = dq2localid
db_dataset = os.environ['ATLAS_DBRELEASE']

dq2=DQ2()
db_locations = dq2.listDatasetReplicas(db_dataset).values()[0][1]

print "db_locations = ", db_locations

db_site = dq2localid
for sitename in TiersOfATLAS.getAllSources():
    if TiersOfATLAS.getSiteProperty(sitename,'alternateName')==dq2localid_alternatename and sitename in db_locations:
        db_site = sitename

outFile = open('db_dq2localid.txt','w')
outFile.write('%s\n' % db_site)
outFile.close()



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: File.py,v 1.2 2008-09-09 14:37:16 moscicki Exp $
##########################################################################

from Ganga.Core.exceptions import GangaException
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.GPIDev.Base.Proxy import isType
from Ganga.GPIDev.Base.Proxy import stripProxy, GPIProxyObjectFactory
import os
import shutil
import uuid

from Ganga.Utility.files import expandfilename, chmod_executable, is_executable

from Ganga.Utility.logging import getLogger

from Ganga.GPIDev.Base.Filters import allComponentFilters

import re

import Ganga.Utility.Config

from Ganga.GPIDev.Lib.File import getSharedPath

# regex [[PROTOCOL:][SETYPE:]..[<alfanumeric>:][/]]/filename
urlprefix = re.compile('^(([a-zA-Z_][\w]*:)+/?)?/')

logger = getLogger()


class File(GangaObject):

    """Represent the files, both local and remote and provide an interface to transparently get access to them.

    Typically in the context of job submission, the files are copied to the directory where the application
    runs on the worker node. The 'subdir' attribute influances the destination directory. The 'subdir' feature
    is not universally supported however and needs a review.

    """
    _schema = Schema(Version(1, 1), {'name': SimpleItem(defvalue="", doc='path to the file source'),
                                     'subdir': SimpleItem(defvalue=os.curdir, doc='destination subdirectory (a relative path)'),
                                     'executable': SimpleItem(defvalue=False, hidden=True, transient=True, doc='specify if executable bit should be set when the file is created (internal framework use)')})
    _category = 'files'
    _name = "File"
    _exportmethods = ["getPathInSandbox", "exists", "create", "isExecutable"]

    # added a subdirectory to the File object. The default is os.curdir, that is "." in Unix.
    # The subdir is a relative path and will be appended to the pathname when writing out files.
    # Therefore changing subdir to a anything starting with "/" will still end up relative
    # to the pathname when the file is copied.
    #
    # There is no protection on putting the parent directory. So ".." is legal and will make
    # the file end up in the parent directory. - AM
    def __init__(self, name=None, subdir=os.curdir):
        super(File, self).__init__()

        if not name is None:
            assert(isinstance(name, str))
            self.name = name

        if not subdir is None:
            self.subdir = subdir

    def __construct__(self, args):
        if len(args) == 1 and isinstance(args[0], str):
            v = args[0]
            import os.path
            expanded = expandfilename(v)
            # if it is not already an absolute filename
            if not urlprefix.match(expanded):
                if os.path.exists(os.path.abspath(expanded)):
                    self.name = os.path.abspath(expanded)
                else:
                    self.name = v
            else:  # bugfix #20545
                self.name = expanded
        else:
            super(File, self).__construct__(args)

    def _attribute_filter__set__(self, attribName, attribValue):
        if attribName is 'name':
            return expandfilename(attribValue)
        return attribValue

    def getPathInSandbox(self):
        """return a relative location of a file in a sandbox: subdir/name"""
        from Ganga.Utility.files import real_basename
        return os.path.join(self.subdir, real_basename(self.name))

    def exists(self):
        """check if the file exists (as specified by 'name')"""
        import os.path
        return os.path.isfile(expandfilename(self.name))

    def create(self, outname):
        """create a file in  a local filesystem as 'outname', maintain
        the original permissions """
        import shutil

        shutil.copy(expandfilename(self.name), outname)
        if self.executable:
            chmod_executable(outname)

    def __repr__(self):
        """Get   the  representation   of  the   file.  Since   the  a
        SimpleStreamer uses  __repr__ for persistency  it is important
        to return  a valid python expression  which fully reconstructs
        the object.  """

        return "File(name='%s',subdir='%s')" % (self.name, self.subdir)

    def isExecutable(self):
        """  return true  if  a file  is  create()'ed with  executable
        permissions,  i.e. the  permissions of  the  existing 'source'
        file are checked"""
        return self.executable or is_executable(expandfilename(self.name))

# add File objects to the configuration scope (i.e. it will be possible to
# write instatiate File() objects via config file)
Ganga.Utility.Config.config_scope['File'] = File

def string_file_shortcut_file(v, item):
    if isinstance(v, str):
        # use proxy class to enable all user conversions on the value itself
        # but return the implementation object (not proxy)
        return stripProxy(File._proxyClass(v))
    return None

allComponentFilters['files'] = string_file_shortcut_file

class ShareDir(GangaObject):

    """Represents the directory used to store resources that are shared amongst multiple Ganga objects.

    Currently this is only used in the context of the prepare() method for certain applications, such as
    the Executable() application. A single ("prepared") application can be associated to multiple jobs.

    """
    _schema = Schema(Version(1, 0), {'name': SimpleItem(defvalue='', doc='path to the file source'),
                                     'subdir': SimpleItem(defvalue=os.curdir, doc='destination subdirectory (a relative path)')})

    _category = 'shareddirs'
    _exportmethods = ['add', 'ls']
    _name = "ShareDir"
    _data = None
#    def _readonly(self):
#        return True

    def __init__(self, name=None, subdir=os.curdir):
        super(ShareDir, self).__init__()
        self._setRegistry(None)

        if not name is None:
            self.name = name
        else:
            # continue generating directory names until we create a unique one
            # (which will likely be on the first attempt).
            while True:
                name = 'conf-{0}'.format(uuid.uuid4())
                if not os.path.isdir(os.path.join(getSharedPath(), name)):
                    os.makedirs(os.path.join(getSharedPath(), name))

                if not os.path.isdir(os.path.join(getSharedPath(), name)):
                    logger.error("ERROR creating path: %s" %
                                 os.path.join(getSharedPath(), name))
                    raise GangaException("ShareDir ERROR")
                else:
                    break
            self.name = str(name)

            # incrementing then decrementing the shareref counter has the effect of putting the newly
            # created ShareDir into the shareref table. This is desirable if a ShareDir is created in isolation,
            # filled with files, then assigned to an application.
            #a=Job(); s=ShareDir(); a.application.is_prepared=s
        #shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
        # shareref.increase(self.name)
        # shareref.decrease(self.name)

    def __deepcopy__(self, memo):
        return super(ShareDir, self).__deepcopy__(memo)

    def add(self, input):
        from Ganga.Core.GangaRepository import getRegistry
        if not isType(input, list):
            input = [input]
        for item in input:
            if isType(item, str):
                if os.path.isfile(expandfilename(item)):
                    logger.info('Copying file %s to shared directory %s' % (item, self.name))
                    shutil.copy2(expandfilename(item), os.path.join(getSharedPath(), self.name))
                    shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                    shareref.increase(self.name)
                    shareref.decrease(self.name)
                else:
                    logger.error('File %s not found' % expandfilename(item))
            elif isType(item, File) and item.name is not '' and os.path.isfile(expandfilename(item.name)):
                logger.info('Copying file object %s to shared directory %s' % (item.name, self.name))
                shutil.copy2(expandfilename(item.name), os.path.join(getSharedPath(), self.name))
                shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                shareref.increase(self.name)
                shareref.decrease(self.name)
            else:
                logger.error('File %s not found' % expandfilename(item.name))

    def ls(self):
        """
        Print the contents of the ShareDir
        """
        full_shareddir_path = os.path.join(getSharedPath(), self.name)
        try:
            os.path.isdir(full_shareddir_path)
            cmd = "find '%s'" % (full_shareddir_path)
            files = os.popen(cmd).read().strip().split('\n')
            padding = '|  '
            for file in files:
                level = file.count(os.sep)
                level = level - 6
                pieces = file.split(os.sep)
                symbol = {0: '', 1: '/'}[os.path.isdir(file)]
                logger.info(padding * level + pieces[-1] + symbol)
        except IOError:
            logger.warn('ShareDir %s not found on storage' %
                        full_shareddir_path)

    def exists(self):
        """check if the file exists (as specified by 'name')"""
        import os.path
        return os.path.isdir(expandfilename(self.name))

    def create(self, outname):
        """create a file in  a local filesystem as 'outname', maintain
        the original permissions """
        import shutil

        shutil.copy(expandfilename(self.name), outname)
        if self.executable:
            chmod_executable(outname)

    def __repr__(self):
        """Get   the  representation   of  the   file.  Since   the  a
        SimpleStreamer uses  __repr__ for persistency  it is important
        to return  a valid python expression  which fully reconstructs
        the object.  """

        return "ShareDir(name='%s',subdir='%s')" % (self.name, self.subdir)

    def isExecutable(self):
        """  return true  if  a file  is  create()'ed with  executable
        permissions,  i.e. the  permissions of  the  existing 'source'
        file are checked"""
        return self.executable or is_executable(expandfilename(self.name))

Ganga.Utility.Config.config_scope['ShareDir'] = ShareDir

def string_sharedfile_shortcut(v, item):
    if isinstance(v, str):
        # use proxy class to enable all user conversions on the value itself
        # but return the implementation object (not proxy)
        return stripProxy(ShareDir._proxyClass(v))
    return None

allComponentFilters['shareddirs'] = string_sharedfile_shortcut
#
#
# $Log: not supported by cvs2svn $
# Revision 1.1  2008/07/17 16:40:53  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.15  2007/08/24 15:55:03  moscicki
# added executable flag to the file, ganga will set the executable mode of the app.exe file (in the sandbox only, the original file is not touched), this is to solve feature request #24452
#
# Revision 1.14  2007/01/25 16:18:21  moscicki
# mergefrom_Ganga-4-2-2-bugfix-branch_25Jan07 (GangaBase-4-14)
#
# Revision 1.12.2.4  2006/11/27 09:30:10  amuraru
# Extended the absolute paths detection in File objects
#
# Revision 1.12.2.3  2006/10/27 15:34:59  amuraru
# *** empty log message ***
#
# Revision 1.12.2.2  2006/10/27 15:33:59  amuraru
# bugs #20545
#
# Revision 1.12  2006/08/29 12:45:08  moscicki
# automatic conversion to absolute path
#
# Revision 1.11  2006/07/27 20:10:45  moscicki
# File to the configuartion symbol scope
#
# Revision 1.10  2006/02/17 14:21:32  moscicki
# fixed exists() method: using expanded file name
#
# Revision 1.9  2006/02/10 14:21:56  moscicki
# added exists() method
# docstring cleanup
#
# Revision 1.8  2005/08/23 17:07:23  moscicki
# Added executable flag for FileBuffer.
# Added create method for File and FileBuffer.
# Added getPathInSandbox() method.
#
# Revision 1.7  2005/08/10 09:45:36  andrew
# Added a subdir to File and FileBuffer objects. Changed the writefile method
# in FileWorspace to use the subdirectory
#
#
#



from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import DQ2Dataset
from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Tasks import Task
from MultiTransform import MultiTransform
from Ganga.Utility.ColourText import overview_colours, ANSIMarkup, fgcol, Effects
markup = ANSIMarkup()
fx = Effects()

from Ganga.Core.exceptions import ApplicationConfigurationError
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import dq2_lock, dq2

from GangaAtlas.Lib.Credentials.ProxyHelper import getNickname 
from dq2.clientapi.DQ2 import DQ2, DQUnknownDatasetException, DQDatasetExistsException, DQFileExistsInDatasetException, DQInvalidRequestException
from dq2.container.exceptions import DQContainerAlreadyHasDataset, DQContainerDoesNotHaveDataset
from dq2.common.DQException import DQException
from Ganga.GPIDev.Schema import *

import copy

o = [""]
def c(s):
   return markup(s,fgcol("blue"))
o.append(markup(" *** Task for Multiple Athena Analyses ***", fgcol("blue")))
#o.append(" Analysis Application     : "+c("t.analysis.application"))
#o.append(" Set Dataset              : "+c("t.setDataset('my.dataset')"))
#o.append(" Input Dataset Object     : "+c("t.analysis.inputdata"))
#o.append(" Output Dataset Object    : "+c("t.analysis.outputdata"))
#o.append(" Files processed per job  : "+c("t.analysis.files_per_job = 10"))
#o.append("")
#o.append(markup("Procedure to do a usual analysis:", fgcol("red")))
##o.append("config.Tasks.merged_files_per_job = 1 # default files per job for merged datasets")
##o.append("config.Tasks.recon_files_per_job = 10 # default files per job for recon (non-merged) datasets")
#o.append("t = AnaTask()")
#o.append('t.name = "MyAnalysisR1"')
#o.append("t.analysis.outputdata.outputdata  = ['nTuple.root' ]")
#o.append("t.analysis.application.option_file = ['./myTopOptions.py' ]")
#o.append("t.analysis.application.prepare()")
#o.append("t.float = 10")
#o.append('t.initializeFromDatasets(["user08.MyName.ganga.dataset.recon.AOD"])')
#o.append("#t.info() # Check here if settings are correct")
#o.append("t.run()")
#o.append("t.overview() # Watch the processing")
#o.append("")
#o.append("A container dataset with all outputs will be created as")
#o.append(c("user.YourNickname.<YYYYMMDDHHMMSS>.task_<id>.<task name>/"))
#o.append("Subtask (Transform) outputs will be put into these datasets:")
#o.append(c("user.YourNickname.<YYYYMMDDHHMMSS>.task_<id>.subtask_<nr>.<inputdataset name>"))
task_help = "\n".join(o)
task_help_nocolor = task_help.replace(fgcol("blue"),"").replace(fx.normal, "").replace(fgcol("red"),"")

from dq2.clientapi.DQ2 import DQ2, DQUnknownDatasetException
dq2=DQ2(force_backend='rucio')

#config.addOption('merged_files_per_job',1,'OBSOLETE', type=int)
#config.addOption('recon_files_per_job',10,'OBSOLETE', type=int)

class MultiTask(Task):
   __doc__ = task_help_nocolor
   _schema = Schema(Version(1,1), dict(Task._schema.datadict.items() + {
       }.items()))
   _category = 'tasks'
   _name = 'MultiTask'
   _exportmethods = Task._exportmethods + ["initializeFromTagInfo", "initializeFromDatasets", "unitOverview", 'getTransform', 'getContainerName', 'listAllDatasets', 'restartTask', 'clearExcludedSites', 'activateAllUnits', 'setAllRunlimits', 'checkContainerContents']
   
   def initialize(self):
      super(MultiTask, self).initialize()
      self.transforms = []

   def getTransform(self, trf):
      """Get transform using either index or name"""
      if isinstance(trf, str):
         for trfid in range(0, len(self.transforms)):
            if trf == self.transforms[trfid].name:
               return self.transforms[trfid]
         logger.warning("Couldn't find transform with name '%s'." % trf)
      elif isinstance(trf, int):
         if trf < 0 and trf > len(self.transforms):
            logger.warning("Transform number '%d' out of range" % trf)
         else:
            return self.transforms[trf]
      else:
         logger.warning('Incorrect type for transform referral. Allowed types are int or string.')

      return None

   def restartTask(self):
      # loop over transforms and units and change status as required
      for trf in self.transforms:
         trf_complete = True
         if trf.status == "pause":
            continue
         
         for uind in range(0, len(trf.unit_partition_list)):
            if not trf.isUnitComplete(uind):
               trf_complete = False
               break
            
         if not trf_complete:
            trf.status = "running"
            self.status = "running"
            
   def getContainerName(self):
      if self.name == "":
         name = "task"
      else:
         name = self.name
         
      name_base = ["user",getNickname(),self.creation_date, name, "id_%i" % self.id ]
      
      return (".".join(name_base) + "/").replace(" ", "_")

   def listAllDatasets(self):
      "List all datasets in container of this transform"
      ds_list = []
      try:
         try:
            dq2_lock.acquire()
            ds_list = dq2.listDatasetsInContainer(self.getContainerName())
         except DQContainerDoesNotHaveDataset:
            pass
         except Exception as x:
            logger.error('Problem finding datasets associated with TRF container %s: %s %s' %( self.getContainerName(), x.__class__, x))
         except DQException as x:
            logger.error('DQ2 Problem finding datasets associated with TRF container %s: %s %s' %( self.getContainerName(), x.__class__, x))
      finally:
          dq2_lock.release()
          
      return ds_list

   def initializeFromDatasets(self,dataset_list):
      """ For each dataset in the dataset_list a unit is created. 
          The output dataset names are set using the run numbers and tags of the input datasets appended to the current t.analysis.outputdata.datasetname field."""
      if not type(dataset_list) is list:
         logger.error("dataset_list must be a python list: ['ds1','ds2',...]")
         return

      # check for primary transforms
      primary_tfs = []
      for tf in self.transforms:
         if len(tf.required_trfs) == 0:
            primary_tfs.append( tf )

      if len(primary_tfs) == 0:
         logger.error("No primary transforms specified. Yout need at least one before the Task can be initialised.")
         return
      
      unit_num = 0
         
      for dset in dataset_list:
         dset = dset.strip()
         try:
            if "*" in dset:
               logger.error("WARNING: Wildcards may include unexpected datasets in your processing! Please list your datasets before specifying them here!")
            try:
               if dset[-1] == "/":
                  tid_datasets = dq2.listDatasetsInContainer(dset)
               else:
                  tid_datasets = dq2.listDatasetsInContainer(dset+"/")
            except DQUnknownDatasetException:
               dslist = dq2.listDatasets(dset).keys()
               if len(dslist) == 0:
                  logger.error("Dataset %s not found!" % dset)
                  return
               tid_datasets = [ds for ds in dslist if "_tid" in ds and not "_sub" in ds]
               if len(tid_datasets) == 0:
                  if len(dslist) > 1:
                     logger.error("Found no tid dataset but multiple datasets match %s*!" % dset)
                     return
                  tid_datasets = [dslist[0]]
         except Exception as e:
            logger.error('DQ2 Error while listing dataset %s*! %s' % (dset, e))
            return
         logger.info("Found %i datasets matching %s..." % (len(tid_datasets), dset))

         if len(tid_datasets) == 0:
            logger.error("No tid datasets found from dataset list. Maybe the container '%s' is empty?" % dset)
            return
         
         prev_num = unit_num
         for tf in primary_tfs:
            unit_num = prev_num
            for ds in tid_datasets:
               tf.addUnit("Unit_%d" % unit_num, ds)
               unit_num += 1

   def initializeFromTagInfo(self,tag_info):
      """ For each dataset in the dataset_list a unit is created. 
          The output dataset names are set using the run numbers and tags of the input datasets appended to the current t.analysis.outputdata.datasetname field."""
      if not type(tag_info) is dict:
         logger.error("tag_info is not a dictionary")
         return

      # check for primary transforms
      primary_tfs = []
      for tf in self.transforms:
         if len(tf.required_trfs) == 0:
            primary_tfs.append( tf )

      if len(primary_tfs) == 0:
         logger.error("No primary transforms specified. Your need at least one before the Task can be initialised.")
         return

      for tf in primary_tfs:
         unit_num = 0      

         # find the number of units
         num_refs = 50
         for dsf in tag_info.keys():

            num_units = round( 0.5 + len(tag_info[dsf]['refs']) / num_refs)

            ref_num = 0
            for i in range(0, num_units):
               tf.addUnit("Unit_%d" % unit_num, ":".join([dsf, str(ref_num), str(ref_num + num_refs)]))
               unit_num += 1
               ref_num += num_refs

         # finally copy the tag info data
         tf.inputdata = DQ2Dataset()
         tf.inputdata.tag_info = copy.deepcopy(tag_info)
               
   def startup(self):
      super(MultiTask,self).startup()

   def check(self):
      super(MultiTask,self).check()

   def help(self):
      print task_help

   def overview(self):
      super(MultiTask, self).overview()

   def unitOverview(self, status = ''):
      """Show a overview of the units. Use the 'status' argument to only view units of a particular status. Options are: 'bad', 'hold', 'running', 'completed', 'ready'"""
      if status and not status in ['bad', 'hold', 'running', 'completed', 'ready']:
         logger.error("Not a valid status for unitOverview. Possible options are: 'bad', 'hold', 'running', 'completed', 'ready'.")
         return
      
      print "Colours: " + ", ".join([markup(key, overview_colours[key])
                                     for key in ["hold", "ready", "running", "completed", "bad"]])
      print "Lists the units for each partition and their current status"
      #print "Format: (partition number)[:(number of failed attempts)]"
      print
      print " "* 41 + "Active\tConf'd\tSub'd\tDwnl'd\tMerged\tComp.\tExcep.\tReason"
      for trfid in range(0, len(self.transforms)):
         print "----------------------------------------------------------------------------------------------------------------------"
         print "----   Transform %d:  %s" % (trfid, self.transforms[trfid].name)
         print
         self.transforms[trfid].unit_overview(status)
         print
      
   def run(self):
      self.checkTRFConsistency()
      super(MultiTask, self).run()

   def checkTRFConsistency(self):
      """Check the consistency of the attached transforms"""
      for trf in self.transforms:

         # check primary trfs
         if len(trf.required_trfs) == 0:

            # ensure the number unit input data, output data and partition lists are setup properly
            if len(trf.unit_partition_list) == 0:
               raise ApplicationConfigurationError(None, "Transform %d (%s) is primary but hasn't been initialised with input data" % (trf.getID(), trf.name))

         else:

            if not trf.isLocalTRF():

               if trf.single_unit:
                  raise ApplicationConfigurationError(None, "Transform %d (%s) set as a single unit but has a grid backend" % (trf.getID(), trf.name))
               
               # this will just take the DQ2 output of the previous TRFs and process each unit itself
               trf.unit_inputdata_list = []
               trf.unit_outputdata_list = []
               trf.unit_state_list = []
               trf.unit_partition_list = []
               
               for req_trf in trf.required_trfs:
                  trf.unit_inputdata_list += [ [] for uind in range(len(self.transforms[req_trf].unit_outputdata_list)) ] 
                  trf.unit_outputdata_list += [ self.transforms[req_trf].unit_outputdata_list[uind] for uind in range(len(self.transforms[req_trf].unit_outputdata_list)) ]
                  trf.unit_partition_list += [ [] for uind in range(len(self.transforms[req_trf].unit_outputdata_list)) ]
                  trf.unit_state_list += [ {'active':True, 'configured':False, 'submitted':False, 'download':False, 'merged':False, 'reason':'', 'exceptions' : 0, 'force' : False} for uind in range(len(self.transforms[req_trf].unit_outputdata_list)) ]
            else:
               
               if trf.single_unit:
                  
                  # we only have one unit
                  trf.unit_inputdata_list = [[]]
                  trf.unit_outputdata_list = [[]]
                  trf.unit_partition_list = [[]]
                  trf.unit_state_list = [{'active':True, 'configured':False, 'submitted':False, 'download':False, 'merged':False, 'reason':'', 'exceptions' : 0, 'force' : False} ]
               else:

                  # we have units dependant on mergers, etc.
                  trf.unit_inputdata_list = []
                  trf.unit_outputdata_list = []
                  trf.unit_partition_list = []
                  trf.unit_state_list = []
               
                  for req_trf in trf.required_trfs:
                     if not self.transforms[req_trf].merger:
                        trf.unit_inputdata_list += [ [] for uind in range(len(self.transforms[req_trf].unit_outputdata_list)) ] 
                        trf.unit_outputdata_list += [ self.transforms[req_trf].unit_outputdata_list[uind] for uind in range(len(self.transforms[req_trf].unit_outputdata_list)) ]
                        trf.unit_partition_list += [ [] for uind in range(len(self.transforms[req_trf].unit_outputdata_list)) ]
                        trf.unit_state_list += [ {'active':True, 'configured':False, 'submitted':False, 'download':False, 'merged':False, 'reason':'', 'exceptions' : 0, 'force' : False} for uind in range(len(self.transforms[req_trf].unit_outputdata_list)) ]
                        
                     else:
                        trf.unit_inputdata_list += [[]]
                        trf.unit_outputdata_list += [[]]
                        trf.unit_partition_list += [[]]
                        trf.unit_state_list += [{'active':True, 'configured':False, 'submitted':False, 'download':False, 'merged':False, 'reason':'', 'exceptions' : 0, 'force' : False} ]

   def clearExcludedSites(self):
      for trf in self.transforms:     
         if hasattr(trf.backend, 'requirements') and hasattr(trf.backend.requirements, 'excluded_sites'):
            trf.backend.requirements.excluded_sites = []  

   def activateAllUnits(self):
      for trf in self.transforms:
         if isinstance(trf,MultiTransform):
            for uind in xrange(0,trf.getNumUnits()):                                    
               trf.activateUnit(uind)   

   def setAllRunlimits(self, runLimit):
      for trf in self.transforms:
         if hasattr(trf, 'setRunlimit'):
            trf.setRunlimit(runLimit)           

   def checkContainerContents(self):
      for trf in self.transforms: 
         if isinstance(trf,MultiTransform):
            trf.checkContainerContents() 
         



from __future__ import absolute_import
##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Proxy.py,v 1.2.4.3 2009-07-10 11:29:27 ebke Exp $
##########################################################################

import Ganga.Utility.logging
from Ganga.Utility.Config import getConfig

from Ganga.GPIDev.Schema import ComponentItem

from Ganga.GPIDev.Base.Objects import Node, GangaObject, ObjectMetaclass, _getName
from Ganga.Core import GangaAttributeError, ProtectedAttributeError, ReadOnlyObjectError, TypeMismatchError

import os

from inspect import isclass

import types

implRef = '_impl'
proxyClass = '_proxyClass'
proxyObject = '_proxyObject'

prepconfig = getConfig('Preparable')

logger = Ganga.Utility.logging.getLogger(modulename=1)

# some proxy related convieniance methods

_knownLists = None

def getRuntimeGPIObject(obj_name, silent=False):
    import Ganga.GPI
    if obj_name in Ganga.GPI.__dict__.keys():
        return Ganga.GPI.__dict__[obj_name]()
    else:
        returnable = raw_eval(obj_name)
        if returnable == obj_name:
            if silent is False:
                logger.error("Cannot find Object: '%s' in GPI. Returning None." % str(obj_name))
                #logger.error("dict: %s" % Ganga.GPI.__dict__)
                #import traceback
                #traceback.print_stack()
            return None
        return returnable

def runtimeEvalString(this_obj, attr_name, val):

    ## Don't check or try to auto-eval non-string objects
    if type(val) != str:
        return val

    raw_obj = stripProxy(this_obj)
    shouldEval = None

    ## Lets check the Schema to see if a string object is allowed here
    ## If this is a ComponentItem we know in advance that we need to try and evaluate this
    ## If the object is NOT a ComponentItem but is still in the schema, check to see if the object is allowed to be a string or not
    ## If an object is NOT allowed to be a string then it should be 'eval'-ed
    if hasattr(raw_obj, '_schema'):

        if not raw_obj._schema.hasAttribute(attr_name):
            shouldEval = True
        else:
            this_attr = raw_obj._schema.getItem(attr_name)
            if isType(this_attr, ComponentItem):
                ## This is a component Item and isn't equivalent to a string
                shouldEval = True
            else:
                allowedTypes = this_attr.getProperties()['typelist']
                for this_type in allowedTypes:
                    if this_type == str:
                        ## This type is a string and shouldn't be evaluated
                        shouldEval = False
                        break
                    else:
                        ## This type is written as a string so need to work out what it is
                        if type(this_type) == str:
                            try:
                                import Ganga.GPI
                                eval_type = eval(this_type, Ganga.GPI.__dict__)
                                if eval_type == str:
                                    ## This type was written as "str" ... slightly annoying but OK...
                                    shouldEval = False
                                    break
                                else:
                                    ## This type is NOT a string so based on this we should Eval
                                    shouldEval = True
                            except Exception as err:
                                logger.debug("Failed to evalute type: %s" % str(this_type))
                                logger.debug("Err: %s" % str(err))
                                ## We can't eval in this case. It may just be the type which is broken
                                shouldEval = True
                        else:
                            ## This type isn't in a string format so we don't need to work out what it is
                            shouldEval = True

    ## If the attribute is not in the Schema lets see if this class instance knows about this object or not
    ## If the attribute is NOT a string but is in this instane then we should try and eval
    ## But if it's known by the instance and is a string, we should just use the value
    elif hasattr(raw_obj, attr_name):

        if type(getattr(raw_obj, attr_name)) == str:
            shouldEval = False
        else:
            shouldEval = True

    ## If the object is not in the schema then try and eval the object anyway as we a-priori don't know any better
    ## THIS IS POTENTIALLY DANGEROUS AND IF A LOT OF USERS COMPLAIN THIS SHOULD BE REVERSED TO FALSE!!!
    else:
        shouldEval = True

    assert(shouldEval is not None)

    if shouldEval is True:
        new_val = raw_eval(val)
    else:
        new_val = val

    return new_val



def raw_eval(val):
    try:
        import Ganga.GPI
        temp_val = eval(val, Ganga.GPI.__dict__)
        if isclass(temp_val):
            new_val = temp_val()
        else:
            new_val = temp_val
    except Exception as err:
        logger.debug("Proxy Cannot evaluate v=: '%s'" % str(val))
        logger.debug("Using raw value instead")
        new_val = val

    if hasattr(stripProxy(new_val), '_auto__init__'):
        stripProxy(new_val)._auto__init__()

    return new_val

def getKnownLists():
    global _knownLists
    if _knownLists is None:
        try:
            from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
        except ImportError:
            _knownLists = None
            return (tuple, list)
        _knownLists = (tuple, list, GangaList)
    return _knownLists

def isProxy(obj):
    """Checks if an object is a proxy"""
#    return isinstance(obj.__class__, GPIProxyObject)
    # Alex changed for below as doesn't check class objects, only instances
    # e.g. isProxy(DiracFile) fails at the Ganga prompt
    if isclass(obj):
        return issubclass(obj, GPIProxyObject) or hasattr(obj, implRef)
    else:
        obj_class = obj.__class__
        return issubclass(obj_class, GPIProxyObject) or hasattr(obj_class, implRef)

def isType(_obj, type_or_seq):
    """Checks whether on object is of the specified type, stripping proxies as needed."""

    obj = stripProxy(_obj)

    bare_type_or_seq = stripProxy(type_or_seq)

    ## Here to avoid circular GangaObject dependencies
    ## is type_or_seq iterable?
    if isinstance(type_or_seq, getKnownLists()):
        clean_list = []
        for type_obj in type_or_seq:
            str_type = type('')
            if type_obj != str_type and type(type_obj) != type(str_type) and (not isclass(type_obj)):
                clean_list.append(type(stripProxy(type_obj)))
            elif isclass(type_obj):
                clean_list.append(type_obj)
            else:
                clean_list.append(type_obj)

        return isinstance(obj, tuple(clean_list))

    else:
        return isinstance(obj, bare_type_or_seq)

def getName(_obj):
    """Strip any proxy and then return an objects name"""
    obj = stripProxy(_obj)
    returnable = _getName(obj)
    return returnable

def stripProxy(obj):
    """Removes the proxy if there is one"""
    return getattr(obj, implRef, obj)


def addProxy(obj):
    """Adds a proxy to a GangaObject"""
    if isType(obj, GangaObject):
        if not isProxy(obj):
            if hasattr(obj, proxyObject):
                return getattr(obj, proxyObject)
            else:
                return GPIProxyObjectFactory(obj)
    return obj


def getProxyAttr(obj, attr_name):
    """Gets an attribute from a proxied object"""
    return getattr(stripProxy(obj), attr_name)


def runProxyMethod(obj, method_name, *args):
    """Calls a method on the object, removing the proxy if needed"""
    fp = getProxyAttr(obj, method_name)
    return fp(*args)


def export(method):
    """
    Decorate a GangaObject method to be exported to the GPI
    """
    method.exported = True
    return method

# apply object conversion or if it fails, strip the proxy and extract the
# object implementation


def stripComponentObject(v, cfilter, item):

    def getImpl(v):
        if v is None:
            if not item['optional']:
                raise TypeMismatchError(None, 'component(%s) is mandatory and None may not be used' % str(getName(item)))
                return v
            else:
                return None
        if isType(v, GangaObject):
            return v
        if not isinstance(v, GPIProxyObject):
            raise TypeMismatchError("cannot assign value '%s', expected a '%s' object " % (repr(v), item['category']))
        return stripProxy(v)

    vv = cfilter(v, item)
    if vv is None:
        return getImpl(v)
    else:
        return vv

from Ganga.GPIDev.TypeCheck import _valueTypeAllowed
valueTypeAllowed = lambda val, valTypeList: _valueTypeAllowed(val, valTypeList, logger)


class ProxyDataDescriptor(object):

    def __init__(self, name):
        self._name = name

    # apply object conversion or if it failes, make the wrapper proxy
    def disguiseComponentObject(self, v):
        # get the proxy for implementation object
        def getProxy(v):
            if not isType(v, GangaObject):
                raise GangaAttributeError("invalid type: cannot assign '%s' to attribute '%s'" % (repr(v), getName(self)))
            return GPIProxyObjectFactory(v)

        # convert implementation object to GPI value according to the
        # static method defined in the implementation object
        def object2value(v):
            return None

        vv = object2value(v)
        if vv is None:
            # allow None to be a legal value for optional component items
            if v is None:
                return None
            else:
                return getProxy(v)

        else:
            return vv

    # apply attribute conversion
    def disguiseAttribute(self, v):
        if isType(v, GangaObject):
            return GPIProxyObjectFactory(v)
        return v

    def __get__(self, obj, cls):

        # at class level return a helper object (for textual description)
        if obj is None:
            # return Schema.make_helper(getattr(getattr(cls, implRef), getName(self)))
            return getattr(stripProxy(cls), getName(self))

        try:
            val = getattr(stripProxy(obj), getName(self))
        except Exception as err:
            if getName(self) in stripProxy(obj).__dict__.keys():
                val = stripProxy(obj).__dict__[getName(self)]
            else:
                val = getattr(stripProxy(obj), getName(self))

        # wrap proxy
        item = stripProxy(obj)._schema[getName(self)]

        if item['proxy_get']:
            return getattr(stripProxy(obj), item['proxy_get'])()

        if isType(item, ComponentItem):
            disguiser = self.disguiseComponentObject
        else:
            disguiser = self.disguiseAttribute

        ## FIXME Add GangaList?
        if item['sequence'] and isType(val, list):
            from Ganga.GPIDev.Lib.GangaList.GangaList import makeGangaList
            val = makeGangaList(val, disguiser)

        returnable = disguiser(val)
        

        if isType(returnable, GangaObject):
            return addProxy(returnable)
        else:
            return returnable


    def _check_type(self, obj, val):
        item = stripProxy(obj)._schema[getName(self)]
        return item._check_type(val, getName(self))

    # apply attribute conversion
    @staticmethod
    def _stripAttribute(obj, v, name):
        # just warn
        # print '**** checking',v,v.__class__,
        # isinstance(val,GPIProxyObject)
        new_v = None
        if isinstance(v, list):
            from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
            v_new = GangaList()
            for elem in v:
                v_new.append(elem)
        if isinstance(v, GPIProxyObject) or hasattr(v, implRef):
            new_v = stripProxy(v)
            logger.debug('%s property: assigned a component object (%s used)' % (name, implRef))

        if new_v is None:
            new_v = v
        return stripProxy(obj)._attribute_filter__set__(name, new_v)

    @staticmethod
    def __app_set__(obj, val):

        if not hasattr(obj, 'application') or obj.application is None:
            return

        if hasattr(obj.application, '_is_prepared'):

            #a=Job(); a.prepare(); a.application=Executable()
            if obj.application.is_prepared not in [None, True] and\
                 hasattr(val, 'is_prepared') and val.is_prepared is None:
                 logger.debug('Overwriting a prepared application with one that is unprepared')
                 obj.application.unprepare()

            #a=Job(); b=Executable(); b.prepare(); a.application=b
        elif obj.application.is_prepared is not True:
            if hasattr(val, 'is_prepared'):
                if val.is_prepared not in [None, True]:
                    from Ganga.Core.GangaRepository import getRegistry
                    shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                    logger.debug('Overwriting application with a prepared one')
                    if stripProxy(obj.application) != val:
                        stripProxy(obj.application).unprepare()
                        shareref.increase(val.is_prepared.name)

            # check that the shared directory actually exists before
            # assigning the (prepared) application to a job
            if hasattr(val, 'is_prepared'):
                if val.is_prepared not in [None, True]:
                    if hasattr(val.is_prepared, 'name'):
                        from Ganga.Utility.files import expandfilename
                        Config_conf = getConfig('Configuration')
                        shared_path = os.path.join(expandfilename(Config_conf['gangadir']), 'shared', Config_conf['user'])
                        if not os.path.isdir(os.path.join(shared_path, val.is_prepared.name)):
                            logger.error('ShareDir directory not found: %s' % val.is_prepared.name)

    @staticmethod
    def __prep_set__(obj, val):

        # if we set is_prepared to None in the GPI, that should effectively
        # unprepare the application
        if val is None:
            if stripProxy(obj).is_prepared is not None:
                logger.info('Unpreparing application.')
                stripProxy(obj).unprepare()

        # Replace is_prepared on an application for another ShareDir object
        if hasattr( stripProxy(obj), '_getRegistry'):
            from Ganga.GPIDev.Lib.File import ShareDir
            if stripProxy(obj)._getRegistry() is not None and isType(val, ShareDir):
                logger.debug('Overwriting is_prepared attribute with a ShareDir object')
                # it's safe to unprepare 'not-prepared' applications.
                stripProxy(obj).unprepare()
                from Ganga.Core.GangaRepository import getRegistry
                shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                shareref.increase(val.name)

        if type(val) is str:
            logger.error("Setting string type to 'is_prepared'")
            #import traceback
            #traceback.print_stack()

    @staticmethod
    def __sequence_set__(stripper, obj, val, name):
        item = stripProxy(obj)._schema[name]
        # we need to explicitly check for the list type, because simple
        # values (such as strings) may be iterable
        new_v = None
        from Ganga.GPIDev.Lib.GangaList.GangaList import makeGangaList
        if isType(val, getKnownLists()):

            # create GangaList
            if stripper is not None:
                new_v = makeGangaList(val, stripper)
            else:
                temp_v = ProxyDataDescriptor._stripAttribute(obj, val, name)
                new_v = makeGangaList(temp_v)
        else:
            # val is not iterable
            if item['strict_sequence']:
                raise GangaAttributeError('cannot assign a simple value %s to a strict sequence attribute %s.%s (a list is expected instead)' % (repr(val), getName(obj), name))
            if stripper is not None:
                new_v = makeGangaList(stripper(val))
            else:
                temp_v = ProxyDataDescriptor._stripAttribute(obj, val, name)
                new_v = makeGangaList(temp_v)

        if new_v is None:
            new_v = val
        return new_v

    @staticmethod
    def __preparable_set__(obj, val, name):
        if obj.is_prepared is not None:
            if obj.is_prepared is not True:
                raise ProtectedAttributeError('AttributeError: "%s" attribute belongs to a prepared application and so cannot be modified.\
                                                unprepare() the application or copy the job/application (using j.copy(unprepare=True)) and modify that new instance.' % (name,))

    ## Inspect this given item to determine if it has editable attributes if it has been set as read-only
    ## Curently Unused although may be useful to keep
    @staticmethod
    def __subitems_read_only(obj):
        can_be_modified = []
        for name, item in obj._schema.allItems():
            ## This object inherits from Node therefore likely has a schema too.
            obj_attr = getattr(obj, name)
            if isType(obj_attr, Node):
                can_be_modified.append( ProxyDataDescriptor.__subitems_read_only(obj_attr) )
            else:

                ## This object doesn't inherit from Node and therefore needs to be evaluated
                if item.getProperties()['changable_at_resubmit']:
                    can_be_modified.append( True )
                else:
                    can_be_modified.append( False )
        
        can_modify = False
        for i in can_be_modified:
            can_modify = can_modify or i

        return can_modify

    @staticmethod
    def __recursive_strip(_val):
        ## Strip the proxies recursively for things like nested lists
        if not isinstance(_val, str) and (isType(_val, getKnownLists()) or\
                (hasattr(stripProxy(_val), '__len__') and hasattr(stripProxy(_val), '__getitem__'))):
            val = stripProxy(_val).__class__()
            if type(val) is dict:
                for _key, elem in _val.iteritems():
                    if isType(_key, GangaObject):
                        key = stripProxy(_key)
                    else:
                        key = _key
                    if isType(elem, GangaObject):
                        val[key] = ProxyDataDescriptor.__recursive_strip(stripProxy(elem))
                    else:
                        val[key] = elem
            else:
                for elem in _val:
                    if isType(elem, GangaObject):
                        val.append(ProxyDataDescriptor.__recursive_strip(stripProxy(elem)))
                    else:
                        val.append(elem)
        else:
            val = stripProxy(_val)
        return val

    def __set__(self, obj, _val):
        # self is the attribute we're about to change
        # obj is the object we're about to make the change in
        # val is the value we're setting the attribute to.
        # item is the schema entry of the attribute we're about to change

        ## Try to remove all proxies
        val = ProxyDataDescriptor.__recursive_strip(_val)

        attr_name = getName(self)

        raw_obj = stripProxy(obj)

        new_val = None

        if not raw_obj._schema.hasAttribute(attr_name):
            raise GangaAttributeError("Cannot assign %s, as it is NOT an attribute in the schema for class: %s" % (attr_name, getName(obj))) 

        #logger.debug("__set__")
        item = raw_obj._schema[attr_name]
        if item['protected']:
            raise ProtectedAttributeError('"%s" attribute is protected and cannot be modified' % (attr_name,))
        if stripProxy(obj)._readonly():

            if not item.getProperties()['changable_at_resubmit']:
                raise ReadOnlyObjectError('object %s is read-only and attribute "%s" cannot be modified now' % (repr(obj), attr_name))
            

        # mechanism for locking of preparable attributes
        if item['preparable']:
            ## Does not modify val
            self.__preparable_set__(obj, val, attr_name)

        # if we set is_prepared to None in the GPI, that should effectively
        # unprepare the application
        if getName(self) == 'is_prepared':
            # Replace is_prepared on an application for another ShareDir object
            ## Does not modify val
            self.__prep_set__(obj, val)

        # catch assignment of 'something'  to a preparable application
        if getName(self) == 'application':
            ## Does not modify val
            self.__app_set__(obj, val)

        # unwrap proxy
        if item.isA(ComponentItem):
            from .Filters import allComponentFilters
            cfilter = allComponentFilters[item['category']]
            stripper = lambda v: stripComponentObject(v, cfilter, item)
        else:
            stripper = None

        if item['sequence']:
            ## Does not modify val
            new_val = self.__sequence_set__(stripper, obj, val, attr_name)
        else:
            if stripper is not None:
                ## Shouldn't modify val
                new_val = stripper(val)
            else:
                ## Does not modify val
                new_val = self._stripAttribute(obj, val, attr_name)

        if new_val is None and val is not None:
            new_val = val

        final_val = None
        # apply attribute filter to component items
        if item.isA(ComponentItem):
            ## Does not modify val
            final_val = self._stripAttribute(obj, new_val, attr_name)
        else:
            final_val = new_val

        if final_val is None and val is not None:
            final_val = val

        ## Does not modify val?
        self._check_type(obj, final_val)

        GangaObject.__setattr__(raw_obj, attr_name, final_val)


class ProxyMethodDescriptor(object):

    def __init__(self, name, internal_name):
        self._name = name
        self._internal_name = internal_name

    def __get__(self, obj, cls):
        if obj is None:
            return getattr(stripProxy(cls), self._internal_name)
        return getattr(stripProxy(obj), self._internal_name)

##########################################################################

# helper to create a wrapper for an existing ganga object

_proxyClassDict={}

def addProxyClass(some_class):
    ## CANNOT USE THE ._name (hence getName) HERE DUE TO REQUIREMENTS OF THE OBJECT IN GPI BEING SANE!!!
    class_name = some_class.__name__
    if class_name not in _proxyClassDict.keys():
        _proxyClassDict[class_name] = GPIProxyClassFactory(class_name, some_class)    
    setattr(some_class, proxyClass, _proxyClassDict[class_name])

def getProxyClass(some_class):
    class_name = some_class.__name__
    if not isclass(some_class):
        from Ganga.Core.exceptions import GangaException
        raise GangaException("Cannot perform getProxyClass on a non-class object: %s:: %s" % (class_name, some_class))
    if not issubclass(some_class, GangaObject):
        from Ganga.Core.exceptions import GangaException
        raise GangaException("Cannot perform getProxyClass on class which is not a subclass of GangaObject: %s:: %s" % (class_name, some_class))
    proxy_class = getattr(some_class, proxyClass, None)
    ## It's possible we ourselves have added a proxy to the base class which we're now inheriting here.
    ## To avoid giving a proxy from Dataset to LHCbDataset and equivalent we'll check against our list of already-found class names.
    if proxy_class is None or class_name not in _proxyClassDict.keys():
        addProxyClass(some_class)
        proxy_class = getattr(some_class, proxyClass)
    return proxy_class

def GPIProxyObjectFactory(_obj):
    # type: (GangaObject) -> GPIProxyObject
    """
    This function _must_ be passed a raw GangaObject. Use :function:`addProxy` for a safe version

    Args:
        _obj (GangaObject): the object to wrap

    Returns:
        a proxy object
    """
    if hasattr(_obj, proxyObject):
        return getattr(_obj, proxyObject)
    if not isType(_obj, GangaObject):
        from Ganga.Core.exceptions import GangaException
        raise GangaException("%s is NOT a Proxyable object" % type(_obj))

    obj_class = _obj.__class__

    proxy_class = getProxyClass(obj_class)

    proxy_object = proxy_class(_proxy_impl_obj_to_wrap=_obj)

    setattr(_obj, proxyObject, proxy_object)

    return proxy_object

# this class serves only as a 'tag' for all generated GPI proxy classes
# so we can test with isinstance rather then relying on more generic but
# less user friendly checking of attribute x._impl


class GPIProxyObject(object):
    pass

# create a new GPI class for a given ganga (plugin) class


def GPIProxyClassFactory(name, pluginclass):
    # type: (str, type(GangaObject)) -> type(GPIProxyObject)
    """
    Args:
        name: the name of the proxy class
        pluginclass: the ``GangaObject`` subclass to wrap

    Returns:
        a new type which wraps ``pluginclass``
    """

    def helptext(f, s):
        if name == '' or name is None:
            _name = ' '
        else:
            _name = name
        f.__doc__ = s % {'classname': _name, 'objname': _name.lower(), 'shortvarname': _name[0].lower()}

    # construct the class on-the-fly using the functions below as methods for
    # the new class

    def _init(self, *args, **kwds):

        ## Zero-th fully initialize self before moving on
        GPIProxyObject.__init__(self)

        ## THE ORDER IN HOW AN OBJECT IS INITIALIZED IS IMPORTANT AND HAS BEEN DOUBLE CHECKED - rcurrie


        ## If we're only constructing a raw Proxy to wrap an existing object lets wrap that and return
        proxy_obj_str = '_proxy_impl_obj_to_wrap'

        if proxy_obj_str in kwds.keys():
            instance = kwds[proxy_obj_str]
            ## Even if we're wrapping something such as here make sure we set all of the proxy related attributes correctly.
            ## Setting of these attributes shold be done here within this class and should probably be properly be done on proxy construction. aka. here
        else:
            ## FIRST INITALIZE A RAW OBJECT INSTANCE CORRESPONDING TO 'pluginclass'
            ## Object was not passed by construction so need to construct new object for internal use
            instance = pluginclass()

        ## Avoid intercepting any of the setter method associated with the implRef as they could trigger loading from disk
        setattr(self, implRef, instance)
        instance.__dict__[proxyObject] = self

        ## Need to avoid any setter methods for GangaObjects
        ## Would be very nice to remove this entirely as I'm not sure a GangaObject should worry about it's proxy (if any)

        if proxy_obj_str in kwds.keys():
            # wrapping not constructing so can exit after determining that the proxy attributes are setup correctly
            return

        ## SECOND WE NEED TO MAKE SURE THAT OBJECT ID IS CORRECT AND THIS DOES THINGS LIKE REGISTER A JOB WITH THE REPO

        instance._auto__init__()

        ## All objects with an _auto__init__ method need to have that method called and we set the various node attributes here based upon the schema
        for key, _val in stripProxy(self)._schema.allItems():
            if not _val['protected'] and not _val['hidden'] and isType(_val, ComponentItem) and key not in Node._ref_list:
                val = stripProxy(getattr(self, key))
                if isinstance(val, GangaObject):
                    val._auto__init__()
                    instance.setNodeAttribute(key, stripProxy(val))
                else:
                    instance.setNodeAttribute(key, stripProxy(val))


        ## THIRD(?) CONSTRUCT THE OBJECT USING THE ARGUMENTS WHICH HAVE BEEN PASSED
        ## e.g. Job(application=exe, name='myJob', ...) or myJob2 = Job(myJob1)
        ## THIS IS PRIMARILY FOR THE 2ND EXAMPLE ABOVE

        ## DOESN'T MAKE SENSE TO KEEP PROXIES HERE AS WE MAY BE PERFORMING A PSEUDO-COPY OP
        clean_args = [stripProxy(arg) for arg in args]
        try:
            stripProxy(self).__construct__(clean_args)
        except TypeError:
            stripProxy(self).__construct__([])


        ## FOURTH ALLOW FOR APPLICATION AND IS_PREPARED etc TO TRIGGER RELAVENT CODE AND SET THE KEYWORDS FROM THE SCHEMA AGAIN
        ## THIS IS MAINLY FOR THE FIRST EXAMPLE ABOVE

        ## THIS CORRECTLY APPLIES A PROXY TO ALL OBJECT ATTRIBUTES OF AN OBJECT CREATED WITHIN THE GPI

        # initialize all properties from keywords of the constructor
        for k in kwds:
            if stripProxy(self)._schema.hasAttribute(k):
                this_arg = stripProxy(kwds[k])
                if hasattr(this_arg, '_auto__init__'):
                    this_arg._auto__init__()

                ## Copying this from the __set__ method in the Proxy descriptor

                if k == 'application':
                    ProxyDataDescriptor.__app_set__(self, this_arg)
                if k == 'is_prepared':
                    ProxyDataDescriptor.__prep_set__(self, this_arg)


                raw_self = stripProxy(self)

                if type(this_arg) is str:
                    this_arg = stripProxy(runtimeEvalString(raw_self, k, this_arg))
                    if hasattr(this_arg, '_auto__init__'):
                        this_arg._auto__init__()

                if type(this_arg) is str:
                    raw_self.setNodeAttribute(k, this_arg)
                    continue
                else:
                    item = pluginclass._schema.getItem(k)

                    # unwrap proxy
                    if item.isA(ComponentItem):
                        from .Filters import allComponentFilters
                        cfilter = allComponentFilters[item['category']]
                        stripper = lambda v: stripComponentObject(v, cfilter, item)
                    else:
                        stripper = None

                    if item['sequence']:
                        this_arg = ProxyDataDescriptor.__sequence_set__(stripper, raw_self, this_arg, k)
                    else:
                        if stripper is not None:
                            this_arg = stripper(this_arg)
                    # apply attribute filter to component items
                    if item.isA(ComponentItem):
                        this_arg = ProxyDataDescriptor._stripAttribute(raw_self, this_arg, k)

                    if hasattr(this_arg, '_auto__init__'):
                        this_arg._auto__init__()

                    raw_self.setNodeAttribute(k, this_arg)
            else:
                logger.warning('keyword argument in the %s constructur ignored: %s=%s (not defined in the schema)', name, k, kwds[k])

        ## end of _init
        return

    from Ganga.Utility.strings import ItemizedTextParagraph

    itbuf = ItemizedTextParagraph('Properties:', linesep='')

    for n, item in pluginclass._schema.allItems():
        if not item['hidden']:
            itbuf.addLine(n, item.describe())

    if not pluginclass.__doc__:
        pluginclass.__doc__ = 'Documentation missing.'

    pluginclass.__doc__.strip()
    pluginclass.__doc__ += "\n\n"

    publicdoc = pluginclass.__doc__ + itbuf.getString()

    helptext(_init, """GPI %(classname)s object constructor:
    %(classname)s() : create %(objname)s with default settings;
    %(classname)s(%(shortvarname)s) : make a copy of %(shortvarname)s;
    %(classname)s(%(shortvarname)s,x=a,...): make a copy of %(shortvarname)s and set property 'x' to a, etc..
    """)

    def _str(self, interactive=False):
        import cStringIO
        sio = cStringIO.StringIO()
        stripProxy(self).printSummaryTree(0, 0, '', out=sio, interactive=interactive)
        returnable = str(sio.getvalue()).rstrip()
        return returnable
    helptext(_str, """Return a printable string representing %(classname)s object as a tree of properties.""")

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text('proxy object...')
            return

        if hasattr(self, implRef):
            raw_self = stripProxy(self)
            if hasattr(raw_self, '_repr_pretty_'):
                raw_self._repr_pretty_(p, cycle)
            elif hasattr(raw_self, '_display'):
                p.text(raw_self._display())
            else:
                #try:
                p.text(self.__str__(interactive=True))
                #except:
                ##    p.text(self.__str__())
        else:
            #try:
            p.text(self.__str__(interactive=True))
            #except:
            #    p.text(self.__str__())

    helptext(_repr_pretty_, """Return a nice string to be printed in the IPython termial""")

    def _repr(self):
        has_proxy = hasattr(self, implRef)
        if has_proxy:
            raw_proxy = stripProxy(self)
        else:
            raw_proxy = None
        if has_proxy and hasattr(raw_proxy, '_repr'):
            return raw_proxy._repr()
        else:
            return '<' + repr(stripProxy(self)) + ' PROXY at ' + hex(abs(id(self))) + '>'
    helptext(_repr, "Return an short representation of %(classname)s object.")

    def _eq(self, x):
        result = False
        if isType(x, GPIProxyObject) or hasattr(x, implRef):
            result = stripProxy(self).__eq__(stripProxy(x))
        else:
            result = stripProxy(self).__eq__(x)
        return result
    helptext(_eq, "Equality operator (==), compare the %(classname)s properties which are declared as [comparable].")

    def _ne(self, x):
        result = True
        if isType(x, GPIProxyObject) or hasattr(x, implRef):
            result = stripProxy(self).__ne__(stripProxy(x))
        else:
            result = stripProxy(self).__ne__(x)
        return result
    helptext(_ne, "Non-equality operator (!=).")

    def _copy(self, unprepare=None):
        logger.debug('unprepare is %s', str(unprepare))
        if unprepare is None:
            if prepconfig['unprepare_on_copy'] is True:
                if hasattr(self, 'is_prepared') or hasattr(self, 'application'):
                    unprepare = True

        def _getSharedPath():
            Config_conf = getConfig('Configuration')
            return os.path.join(expandfilename(Config_conf['gangadir']), 'shared', Config_conf['user'])

        if hasattr(self, 'application'):
            if hasattr(self.application, 'is_prepared'):
                from Ganga.Utility.files import expandfilename
                if self.application.is_prepared not in [None, True]:
                    if hasattr(self.application.is_prepared, 'name'):
                        shared_path = _getSharedPath()
                        if os.path.isdir(os.path.join(shared_path, self.application.is_prepared.name)):
                            from Ganga.Core.GangaRepository import getRegistry
                            shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                            logger.debug('increasing counter from proxy.py')
                            shareref.increase(self.application.is_prepared.name)
                            logger.debug('Found ShareDir directory: %s' % self.application.is_prepared.name)
                elif self.application.is_prepared not in [None, True]:
                    shared_path = _getSharedPath()
                    if not os.path.isdir(os.path.join(shared_path, self.application.is_prepared.name)):
                        logger.error('ShareDir directory not found: %s' % self.application.is_prepared.name)
                        logger.error('Unpreparing Job #%s' % self.id)
                        from Ganga.Core.GangaRepository import getRegistry
                        shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                        shareref.increase(self.application.is_prepared.name)
                        self.unprepare()

        if unprepare is True:
            if hasattr(self, 'is_prepared'):
                from Ganga.Utility.files import expandfilename
                if self.is_prepared not in [None, True]:
                    if hasattr(self.is_prepared, 'name'):
                        shared_path = _getSharedPath()
                        if not os.path.isdir(os.path.join(shared_path, self.is_prepared.name)):
                            logger.error('ShareDir directory not found: %s' % self.is_prepared.name)
                            logger.error('Unpreparing %s application' % getName(stripProxy(self)))
                            self.unprepare()

            c = stripProxy(self).clone()
            if hasattr(c, 'is_prepared') and c._getRegistry() is None:
                from Ganga.Core.GangaRepository import getRegistry
                shareref = GPIProxyObjectFactory(getRegistry("prep").getShareRef())
                shareref.increase(self.is_prepared.name)
            stripProxy(c)._auto__init__(unprepare=True)
        else:
            c = stripProxy(self).clone()
            stripProxy(c)._auto__init__()

        return addProxy(c)

    helptext(_copy, "Make an identical copy of self.")

    def _setattr(self, x, v):
        'This is the setter method in the Proxied Objects'
        #logger.debug("_setattr")
        # need to know about the types that require metadata attribute checking
        # this allows derived types to get same behaviour for free.
        p_Ref = stripProxy(self)
        if p_Ref is not None:
            if not isclass(p_Ref):
                class_type = type(p_Ref)
            else:
                class_type = p_Ref
        else:
            class_type = p_Ref

        if x == implRef and not isinstance(v, class_type):
            raise AttributeError("Internal implementation object '%s' cannot be reassigned" % implRef)

        elif not stripProxy(self)._schema.hasAttribute(x):
            from Ganga.GPIDev.Lib.Job.MetadataDict import MetadataDict
            if hasattr(stripProxy(self), 'metadata') and isType(stripProxy(self).metadata, MetadataDict):
                if x in stripProxy(self).metadata.data.keys():
                    raise GangaAttributeError("Metadata item '%s' cannot be modified" % x)

            if x != implRef:
                raise GangaAttributeError("Can't assign '%s' as it does NOT appear in the object schema for class '%s'" % (x, getName(self)))

        new_v = stripProxy(runtimeEvalString(self, x, v))
        GPIProxyObject.__setattr__(self, x, stripProxy(new_v))


    helptext(_setattr, """Set a property of %(classname)s with consistency and safety checks.
Setting a [protected] or a unexisting property raises AttributeError.""")

    #    def _getattr(self, name):
#        if name == '_impl': return self._impl
#        if '_attribute_filter__get__' in dir(self._impl):
#            return self._impl._attribute_filter__get__(name)
#        return self.name
#        ## need to know about the types that require metadata attribute checking
#        ## this allows derived types to get same behaviour for free.
#        from Ganga.GPIDev.Lib.Job.Job import Job
#        from Ganga.GPIDev.Lib.Tasks.Task import Task
#        from Ganga.GPIDev.Lib.Tasks.Transform import Transform
#        metadata_objects=[Job]
#        if True in (isType(self,t) for t in metadata_objects):
#            try:
#                return self.metadata[name]
#            except:
#                return object.__getattribute__(self,name)
#        return object.__getattribute__(self,name)

    def _getattribute(self, name):

        #logger.debug("_getattribute: %s" % str(name))

        if name.startswith('__') or name == implRef:
            return GPIProxyObject.__getattribute__(self, name)
        else:
            implInstance = stripProxy(self)

            if '_attribute_filter__get__' in dir(implInstance) and \
                    not isType(implInstance, ObjectMetaclass) and \
                    implInstance._schema.hasItem(name) and \
                    not implInstance._schema.getItem(name)['hidden']:
                        returnable = addProxy(implInstance._attribute_filter__get__(name))
            else:
                returnable = GPIProxyObject.__getattribute__(self, name)

        if isType(returnable, GangaObject):
            return addProxy(returnable)
        else:
            return returnable

    # but at the class level _impl is a ganga plugin class
    d = {implRef: pluginclass,
            '__init__': _init,
            '__str__': _str,
            '__repr__': _repr,
            '_repr_pretty_': _repr_pretty_,
            '__eq__': _eq,
            '__ne__': _ne,
            'copy': _copy,
            '__doc__': publicdoc,
            '__setattr__': _setattr,
         #          '__getattr__': _getattr,
            '__getattribute__': _getattribute,
         }

    if not hasattr(pluginclass, '_exportmethods'):
        pluginclass._exportmethods = []

    exported_methods = pluginclass._exportmethods

    # export public methods of this class and also of all the bases
    # this class is scanned last to extract the most up-to-date docstring
    dicts = (b.__dict__ for b in reversed(pluginclass.__mro__))
    for dct in dicts:
        for k in dct:
            if getattr(dct[k], 'exported', False):
                exported_methods.append(k)  # Add all @export'd methods
            if k in exported_methods:
                internal_name = "_export_" + k
                if internal_name not in dct.keys():
                    internal_name = k
                try:
                    method = dct[internal_name]
                except KeyError as err:
                    logger.debug("ObjectMetaClass Error internal_name: %s,\t d: %s" % (internal_name, d))
                    logger.debug("ObjectMetaClass Error: %s" % err)
                    raise err

                if not isinstance(method, types.FunctionType):
                    continue
                f = ProxyMethodDescriptor(k, internal_name)
                f.__doc__ = method.__doc__
                d[k] = f

    # export visible properties... do not export hidden properties
    for attr, item in pluginclass._schema.allItems():
        if not item['hidden']:
            d[attr] = ProxyDataDescriptor(attr)

    def __getitem(self, arg):

        if not hasattr(stripProxy(self), '__getitem__'):
            raise AttributeError('I (%s) do not have a __getitem__ attribute' % str(getName(self)))

        output = stripProxy(self).__getitem__(args)

        if isType(output, GangaObject):
            return addProxy(output)
        else:
            return output

    ## NOT ENABLED YET rcurrie
    #if hasattr(pluginclass, '__getitem__'):
    #    d['__getitem__'] = __getitem
    #d['__getitem__'] = __getitem

    # TODO: this makes GangaList inherit from the list
    # this is not tested and specifically the TestGangaList/testAllListMethodsExported should be verified
    # if name == "GangaList":
    # return type(name, (GPIProxyObject,list), d)

    return type(name, (GPIProxyObject,), d)


#
#
# $Log: not supported by cvs2svn $
# Revision 1.2.4.2  2009/07/08 15:39:15  ebke
# removed object_filter__get__
#
# Revision 1.2.4.1  2009/07/08 11:18:21  ebke
# Initial commit of all - mostly small - modifications due to the new GangaRepository.
# No interface visible to the user is changed
#
# Revision 1.2  2008/09/09 14:37:16  moscicki
# bugfix #40220: Ensure that default values satisfy the declared types in the schema
#
# factored out type checking into schema module, fixed a number of wrongly declared schema items in the core
#
# Revision 1.1  2008/07/17 16:40:52  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.32.4.15  2008/07/02 13:19:35  moscicki
# TODO comment
#
# Revision 1.32.4.14  2008/06/04 10:34:54  moscicki
# fixed typecheck function naming
#
# Revision 1.32.4.13  2008/06/02 11:42:12  moscicki
# fixed typo...
#
# Revision 1.32.4.12  2008/05/27 18:02:50  kuba
# fixed bug #36836, .ganga.log in afs home directory
#
# Revision 1.32.4.11  2008/02/29 09:16:17  moscicki
# fix from Will Reece
#
# Revision 1.32.4.10  2008/02/28 15:45:30  moscicki
# fixed GangaList typechecking problem
#
# Revision 1.32.4.9  2008/02/06 16:26:30  moscicki
# typelist == None => disable type checking
# extra warnings for incomplete typesystem information
#
# Revision 1.32.4.8  2008/02/06 09:48:44  wreece
# Allows the proxy stripper a NoOp if the object has already been stripped.
#
# Revision 1.32.4.7  2008/02/06 09:28:48  wreece
# First pass at a cleanup of the gangalist stuff. I've made changes so the diffs with the 4.4 series are more transparent. Currently still test failures.
#
# Revision 1.32.4.6  2007/12/18 16:40:39  moscicki
# bugfix
#
# Revision 1.32.4.5  2007/12/18 09:05:03  moscicki
# integrated typesystem from Alvin and made more uniform error reporting
#
# Revision 1.32.4.4  2007/11/20 14:29:49  wreece
# Corrects typo in TFile, and removes extra methods from Proxy. Typo was causing tests to fail.
#
# Revision 1.32.4.3  2007/11/14 13:03:54  wreece
# Changes to make shortcuts work correctly with gangalists. all but one tests should now pass.
#
# Revision 1.32.4.2  2007/11/07 17:02:10  moscicki
# merged against Ganga-4-4-0-dev-branch-kuba-slices with a lot of manual merging
#
# Revision 1.32.4.1  2007/11/07 15:10:02  moscicki
# merged in pretty print and GangaList support from ganga-5-dev-branch-4-4-1-will-print branch
#
# Revision 1.32.8.2  2007/10/30 14:30:23  wreece
# Non-working update. Adds in Kuba's exported methods dodge. It is now possible to define a _export_ version of a method for external use and a undecorated method for internal use.
#
# Revision 1.32.8.1  2007/10/30 12:12:08  wreece
# First version of the new print_summary functionality. Lots of changes, but some known limitations. Will address in next version.
#
# Revision 1.32  2007/07/10 13:08:29  moscicki
# docstring updates (ganga devdays)
#
# Revision 1.31  2007/07/10 07:50:52  moscicki
# fixed #27541
#
# Revision 1.30.8.1  2007/06/18 10:16:34  moscicki
# slices prototype
#
# Revision 1.30  2007/03/07 09:24:34  moscicki
# AGAIN: fixed a problem of assigning iterable plain value (such as string) to a non-strict sequence (so x.files = "abc" was yelding 3 files "a","b","c"
#
# added GangaAttributeError exception
#
# Revision 1.29  2007/03/06 16:38:24  moscicki
# fixed a problem of assigning iterable plain value (such as string) to a non-strict sequence (so x.files = "abc" was yelding 3 files "a","b","c"
#
# Revision 1.28  2007/03/05 12:03:01  moscicki
# explicit switch for strict_sequence (default is True), if the sequence is non-strict then a single value v will be converted to [v] on assignment, for example non-strict File sequence yields obj.x = 'a' <=> obj.x = [File('a')]  <=> obj.x = File('a')
#
# Revision 1.27  2007/02/28 18:20:49  moscicki
# moved GangaException to Ganga.Core
#
# Revision 1.26  2006/10/23 11:00:42  moscicki
# minor logger fixed (level changed)
#
# Revision 1.25  2006/09/15 14:19:44  moscicki
# Fixed bug #12229 overview: FutureWarning from Python
#
# Revision 1.24  2006/08/11 13:13:05  adim
# Added: GangaException as a markup base class for all exception that need to be printed in a usable way in IPython shell
#
# Revision 1.23  2005/12/02 15:28:46  moscicki
# customizable _repr() method on GPI objects
#
# Revision 1.22  2005/11/14 13:58:27  moscicki
# fixed a getter filter bug
#
# Revision 1.21  2005/11/14 10:30:37  moscicki
# enabled getter attribute filter
#
# Revision 1.20  2005/09/21 09:09:14  moscicki
# "better" repr text for proxies
#
# Revision 1.19  2005/08/26 13:21:47  moscicki
# do not document hidden properties
#
# Revision 1.18  2005/08/26 10:21:31  karl
# KH: Minor correction: linesep -> separator
#
# Revision 1.17  2005/08/26 09:54:55  moscicki
# minor changes
#
# Revision 1.16  2005/08/24 15:41:19  moscicki
# automatically generated help for properties, disabled the SchemaHelper and few other improvements to the help system
#
# Revision 1.15  2005/08/23 17:15:06  moscicki
# *** empty log message ***
#
#
#



from __future__ import print_function

##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Root.py,v 1.4 2008-09-12 08:08:58 wreece Exp $
##########################################################################

from Ganga.GPIDev.Adapters.IPrepareApp import IPrepareApp
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, FileItem
from Ganga.GPIDev.Lib.File import File, ShareDir

from Ganga.Utility.Config import getConfig, ConfigError
from Ganga.Utility.root import getrootsys, getpythonhome

from Ganga.Core import ApplicationPrepareError

import Ganga.Utility.logging
import inspect
import os
import sys
import tempfile
from Ganga.Utility.files import expandfilename

from Ganga.GPIDev.Base.Proxy import getName

logger = Ganga.Utility.logging.getLogger()
config = getConfig('ROOT')

def getDefaultScript():
    name = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))), 'defaultRootScript.C')
    return File(name)

class Root(IPrepareApp):

    """
    Root application -- running ROOT

    To run a job in ROOT you need to specify the CINT script to be
    executed. Additional files required at run time (shared libraries,
    source files, other scripts, Ntuples) should be placed in the
    inputsandbox of the job. Arguments can be passed onto the script using
    the 'args' field of the application.

    Defining a Simple Job:

    As an example the script analysis.C in the directory ~/abc might
    contain:

    void analysis(const char* type, int events) {
      std::cout << type << "  " << events << std::endl;
    }

    To define an LCG job on the Ganga command line with this script, 
    running in ROOT version 5.14.00b with the arguments 'MinBias' 
    and 10, you would do the following:

    r = Root()
    r.version = '6.04.02'
    r.script = '~/abc/analysis.C'
    r.args = ['Minbias', 10]

    j = Job(application=r, backend=LCG())

    Using Shared Libraries:

    If you have private shared libraries that should be loaded you need to
    include them in the inputsandbox. Files you want back as a result of
    running your job should be placed in your outputsandbox. 

    The shared library mechanism is particularly useful in order to create 
    a thin wrapper around code that uses precompiled libraries, or
    that has not been designed to work in the CINT environment.

    **For more detailed instructions, see the following Wiki page:**

    https://twiki.cern.ch/twiki/bin/view/ArdaGrid/HowToRootJobsSharedObject

    A summary of this page is given below:

    Consider the follow in CINT script, runMain.C, that makes use of a ROOT 
    compatible shared library:

    void runMain(){
      //set up main, eg command line opts
      char* argv[] = {"runMain.C","--muons","100"};
      int argc = 3;

      //load the shared library
      gSystem->Load("libMain");

      //run the code
      Main m(argv,argc);
      int returnCode = m.run();
    }

    The class Main is as follows and has been compiled into a shared
    library, libMain.so. 

    Main.h:

    #ifndef MAIN_H
    #define MAIN_H
    #include "TObject.h"

    class Main : public TObject {

        public:
          Main(){}//needed by Root IO
          Main(char* argv[], int argc);
          int run();

          ClassDef(Main,1)//Needed for CINT
    };
    #endif

    Main.cpp:

    #include <iostream>
    using std::cout;
    using std::endl;
    #include "Main.h"

    ClassImp(Main)//needed for CINT
    Main::Main(char* arvv[], int argc){
      //do some setup, command line opts etc
    }

    int Main::run(){
      cout << "Running Main..." << endl;
      return 0;
    }

    To run this on LCG, a Job could be created as follows:

    r = Root()
    r.version = '5.12.00' #version must be on LCG external site
    r.script = 'runMain.C'

    j = Job(application=r,backend=LCG())
    j.inputsandbox = ['libMain.so']

    It is a requirement that your script contains a function with the same
    name as the script itself and that the shared library file is built to
    be binary compatible with the Grid environment (e.g. same architecture 
    and version of gcc). As shown above, the wrapper class must be made CINT 
    compatible. This restriction does not, however, apply to classes used by 
    the wrapper class. When running remote (e.g. LCG) jobs, the architecture
    used is 'slc3_ia32_gcc323' if the Root version is 5.16 or earlier and
    'slc4_ia32_gcc34' otherwise. This reflects the availability of builds
    on the SPI server:

    http://service-spi.web.cern.ch/service-spi/external/distribution/


    For backends that use a local installation of ROOT the location should
    be set correctly in the [Root] section of the configuration.

    Using Python and Root:

    The Root project provides bindings for Python, the language supported by 
    the Ganga command line interface. These bindings are referred to as PyRoot.
    A job is run using PyRoot if the script has the '.py' extension or the 
    usepython flag is set to True.

    There are many example PyRoot scripts available in the Root tutorials. 
    A short example is given below:

    gengaus.py:

    if __name__ == '__main__':
        from ROOT import gRandom

        output = file('gaus.txt','w')
        try:
            for i in range(100):
                print(gRandom.Gaus(), file=output)
        finally:
            output.close()

    The above script could be run in Ganga as follows:

    r = Root()
    r.version = '5.14.00'
    r.script = '~/gengaus.py'
    r.usepython = True #set automatically for '.py' scripts

    j = Job(application=r,backend=Local())
    j.outputsandbox = ['gaus.txt']
    j.submit()

    When running locally, the python interpreter used for running PyRoot jobs
    will default to the one being used in the current Ganga session.
    The Root binaries selected must be binary compatible with this version.

    The pythonhome variable in the [Root] section of .gangarc controls which
    interpreter will be used for PyRoot jobs.

    When using PyRoot on a remote backend, e.g. LCG, the python version that
    is used will depend on that used to build the Root version requested.

    """
    _schema = Schema(Version(1, 1), {
        'script': FileItem(defvalue=getDefaultScript(), preparable=1, doc='A File object specifying the script to execute when Root starts', checkset='_checkset_script'),
        'args': SimpleItem(defvalue=[], typelist=['str', 'int'], sequence=1, doc="List of arguments for the script. Accepted types are numerics and strings"),
        'version': SimpleItem(defvalue='6.04.02', doc="The version of Root to run"),
        'usepython': SimpleItem(defvalue=False, doc="Execute 'script' using Python. The PyRoot libraries are added to the PYTHONPATH."),
        'is_prepared': SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)', 'bool'], protected=1, hidden=0, comparable=1, doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        'hash': SimpleItem(defvalue=None, typelist=['type(None)', 'str'], hidden=1, doc='MD5 hash of the string representation of applications preparable attributes')
    })
    _category = 'applications'
    _name = 'Root'
    _exportmethods = ['prepare', 'unprepare']

    def __init__(self):
        super(Root, self).__init__()

        from Ganga.GPIDev.Lib.File import getSharedPath

        self.shared_path = Ganga.GPIDev.Lib.File.getSharedPath()

    def configure(self, masterappconfig):
        return (None, None)

    def unprepare(self, force=False):
        """
        Revert a Root() application back to it's unprepared state.
        """
        logger.debug('Running unprepare in Root app')
        if self.is_prepared is not None:
            self.decrementShareCounter(self.is_prepared.name)
            self.is_prepared = None
        self.hash = None

    def prepare(self, force=False):
        """
        A method to place the Root application into a prepared state.
        """
        if (self.is_prepared is not None) and (force is not True):
            raise ApplicationPrepareError(
                '%s application has already been prepared. Use prepare(force=True) to prepare again.' % getName(self))
        self.is_prepared = ShareDir()
        logger.info('Created shared directory: %s' % (self.is_prepared.name))

        try:
            copy_worked = self.copyPreparables()
            if copy_worked == 0:
                logger.error(
                    'Failed during prepare() phase. Unpreparing application.')
                self.unprepare()
                return 0
            else:
                # add the newly created shared directory into the metadata
                # system if the app is associated with a persisted object
                self.checkPreparedHasParent(self)
                self.post_prepare()
                return 1
        except:
            self.unprepare()
            raise

    def _checkset_script(self, value):
        """Callback used to set usepython to 1 if the script name has a *.py or *.PY extention."""
        from os.path import splitext
        (_, ext) = splitext(str(value.name))
        # use pyroot if this is a python script
        if('.py' == ext.lower()):
            logger.debug('Setting usepython to True')
            self.usepython = True


class RootRTHandler(IRuntimeHandler):

    def quoteCintArgString(self, cintArg):
        return '\\"%s\\"' % cintArg

    def _getRootEnvSys(self, version, usepython=False):
        """Returns an environment suitable for running Root and sometimes Python."""
        from os.path import join
        from os import environ

        from Ganga.Lib.Root.shared import setEnvironment, findPythonVersion

        rootsys = getrootsys(version)

        logger.info("rootsys: %s" % str(rootsys))

        rootenv = {}
        # propagate from localhost
        if 'PATH' in environ:
            setEnvironment('PATH', environ['PATH'], update=True, environment=rootenv)
        if 'LD_LIBRARY_PATH' in environ:
            setEnvironment('LD_LIBRARY_PATH', environ['LD_LIBRARY_PATH'], update=True, environment=rootenv)

        setEnvironment('LD_LIBRARY_PATH', join(rootsys, 'lib'), update=True, environment=rootenv)
        setEnvironment('PATH', join(rootsys, 'bin'), update=True, environment=rootenv)
        setEnvironment('ROOTSYS', rootsys, update=False, environment=rootenv)
        logger.debug('Have set Root variables. rootenv is now %s', str(rootenv))

        if usepython:
            # first get from config
            python_version = ''
            try:
                python_version = getConfig('ROOT')['pythonversion']
            except ConfigError as e:
                logger.debug('There was a problem trying to get [ROOT]pythonversion: %s.', e)

            logger.debug('Found version of python: %s', str(python_version))

            # now try grepping files
            if not python_version:
                python_version = findPythonVersion(rootsys)

            if (python_version is None):
                logger.warn('Unable to find the Python version needed for Root version %s. See the [ROOT] section of your .gangarc file.', version)
            else:
                logger.debug('Python version found was %s', python_version)
            python_home = getpythonhome(pythonversion=python_version)
            logger.info("Looking in: %s" % python_home)
            logger.debug('PYTHONHOME is being set to %s', python_home)

            python_bin = join(python_home, 'bin')
            setEnvironment('PATH', python_bin, update=True, environment=rootenv)
            setEnvironment('PYTHONPATH', join(rootsys, 'lib'), update=True, environment=rootenv)
            logger.debug('Added PYTHONPATH. rootenv is now %s', str(rootenv))

            if join(python_bin, 'python') != sys.executable:
                # only try to do all this if the python currently running isn't
                # going to be used
                logger.debug('Using a different Python - %s.', python_home)
                python_lib = join(python_home, 'lib')

                if not os.path.exists(python_bin) or not os.path.exists(python_lib):
                    logger.error('The PYTHONHOME specified does not have the expected structure. See the [ROOT] section of your .gangarc file.')
                    logger.error('PYTHONPATH is: ' + str(os.path))

                setEnvironment('LD_LIBRARY_PATH', python_lib, update=True, environment=rootenv)
                setEnvironment('PYTHONHOME', python_home, update=False, environment=rootenv)
                setEnvironment('PYTHONPATH', python_lib, update=True, environment=rootenv)

        return (rootenv, rootsys)

    def _prepareCintJobConfig(self, app, appconfig, appmasterconfig, jobmasterconfig):
        """JobConfig for executing a Root script using CINT."""
        from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
        from os.path import join, split
        import string

        from Ganga.GPIDev.Lib.File import getSharedPath

        # Arguments to the ROOT script needs to be a comma separated list
        # enclosed in (). Strings should be enclosed in escaped double quotes.
        arglist = []
        for arg in app.args:
            if isinstance(arg, str):
                arglist.append(self.quoteCintArgString(arg))
            else:
                arglist.append(arg)
        rootarg = '\(' + string.join([repr(s) for s in arglist], ',') + '\)'

        script = app.script
        if script == File():
            script = File(defaultScript())
        else:
            script = File(os.path.join(os.path.join(Ganga.GPIDev.Lib.File.getSharedPath(), app.is_prepared.name), os.path.basename(app.script.name)))

        # Start ROOT with the -b and -q options to run without a
        # terminal attached.
        arguments = ['-b', '-q', os.path.relpath(join('.', script.subdir,
                                      split(script.name)[1])) + rootarg]
        inputsandbox = app._getParent().inputsandbox + [script]

        (rootenv, _) = self._getRootEnvSys(app.version)
        logger.debug("ROOT environment:\n %s: ", str(rootenv))

        returnable = StandardJobConfig('root.exe', inputsandbox, arguments,
                                 app._getParent().outputsandbox)

        logger.debug("root jobconfig: %s" % str(returnable))

        return returnable

    def _preparePyRootJobConfig(self, app, appconfig, appmasterconfig, jobmasterconfig):
        """JobConfig for executing a Root script using CINT."""
        from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
        from os.path import join, split

        from Ganga.GPIDev.Lib.File import getSharedPath

        script = app.script
        if script == File():
            script = File(defaultPyRootScript())
        else:
            script = File(os.path.join(os.path.join(Ganga.GPIDev.Lib.File.getSharedPath(), app.is_prepared.name), os.path.basename(app.script.name)))

        arguments = [join('.', script.subdir, split(script.name)[1])]
        arguments.extend([str(s) for s in app.args])
        arguments.append('-b')

        inputsandbox = app._getParent().inputsandbox + [script]

        (rootenv, _) = self._getRootEnvSys(app.version, usepython=True)
        logger.debug("PyRoot environment:\n %s: ", str(rootenv))

        return StandardJobConfig('python', inputsandbox, arguments,
                                 app._getParent().outputsandbox)

    def prepare(self, app, appconfig, appmasterconfig, jobmasterconfig):
        """The default prepare method. Used to select scripting backend."""
        if(app.usepython):
            return self._preparePyRootJobConfig(app, appconfig,
                                                appmasterconfig, jobmasterconfig)
        else:
            return self._prepareCintJobConfig(app, appconfig,
                                              appmasterconfig, jobmasterconfig)


class RootLocalRTHandler(RootRTHandler):

    """Same as RootRTHander, but slight difference in string quoting used."""

    def quoteCintArgString(self, cintArg):
        return '"%s"' % cintArg


class RootDownloadHandler(IRuntimeHandler):

    def prepare(self, app, appconfig, appmasterconfig, jobmasterconfig):
        from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
        runScript, inputsandbox, rootenv = downloadWrapper(app)

        # propage command line args
        argList = [str(s) for s in app.args]

        return StandardJobConfig(runScript, inputsandbox, argList,
                                 app._getParent().outputsandbox)


class RootLCGRTHandler(IRuntimeHandler):

    def prepare(self, app, appconfig, appmasterconfig, jobmasterconfig):
        from Ganga.Lib.LCG import LCGJobConfig

        runScript, inputsandbox, rootenv = downloadWrapper(app)

        # propage command line args
        argList = [str(s) for s in app.args]

        return LCGJobConfig(runScript, inputsandbox, argList,
                            app._getParent().outputsandbox, rootenv)

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('Root', 'Cronus', RootRTHandler)
allHandlers.add('Root', 'LSF', RootLocalRTHandler)
allHandlers.add('Root', 'Local', RootLocalRTHandler)
allHandlers.add('Root', 'Interactive', RootRTHandler)
# Easy way to test the system with ROOT dowloaded on demand
#allHandlers.add('Root','Local', RootDownloadHandler)
allHandlers.add('Root', 'PBS', RootLocalRTHandler)
allHandlers.add('Root', 'SGE', RootLocalRTHandler)
allHandlers.add('Root', 'Condor', RootRTHandler)
allHandlers.add('Root', 'LCG', RootLCGRTHandler)
allHandlers.add('Root', 'CREAM', RootLCGRTHandler)
allHandlers.add('Root', 'ARC', RootLCGRTHandler)
allHandlers.add('Root', 'TestSubmitter', RootRTHandler)
allHandlers.add('Root', 'Remote', RootLCGRTHandler)


def downloadWrapper(app):
    from os.path import join, split
    from Ganga.GPIDev.Lib.File import FileBuffer
    import string

    from Ganga.GPIDev.Lib.File import getSharedPath

    rootsys = join('.', 'root')
    rootenv = {'ROOTSYS': rootsys}

    script = app.script
    if script == File():
        if not app.usepython:
            script = File(defaultScript())
        else:
            script = File(defaultPyRootScript())
    else:
        script = File(os.path.join(os.path.join(Ganga.GPIDev.Lib.File.getSharedPath(), app.is_prepared.name), os.path.basename(app.script.name)))

    commandline = ''
    scriptPath = join('.', script.subdir, split(script.name)[1])
    if not app.usepython:
        # Arguments to the ROOT script needs to be a comma separated list
        # enclosed in (). Strings should be enclosed in escaped double quotes.
        arglist = []
        for arg in app.args:
            if isinstance(arg, str):
                arglist.append('\\\'' + arg + '\\\'')
            else:
                arglist.append(arg)
        rootarg = '\(\"' + string.join([str(s) for s in arglist], ',') + '\"\)'

        # use root
        commandline = 'root.exe -b -q ' + scriptPath + rootarg + ''
    else:
        # use python
        pyarg = string.join([str(s) for s in app.args], ' ')
        commandline = '\'%(PYTHONCMD)s ' + scriptPath + ' ' + pyarg + ' -b \''

    logger.debug("Command line: %s: ", commandline)

    # Write a wrapper script that installs ROOT and runs script
    script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                                                   'wrapperScriptTemplate.py')
    from Ganga.GPIDev.Lib.File import FileUtils
    wrapperscript = FileUtils.loadScript(script_location, '')

    wrapperscript = wrapperscript.replace('###COMMANDLINE###', commandline)
    wrapperscript = wrapperscript.replace('###ROOTVERSION###', app.version)
    wrapperscript = wrapperscript.replace('###SCRIPTPATH###', scriptPath)
    wrapperscript = wrapperscript.replace('###USEPYTHON###', str(app.usepython))

    logger.debug('Script to run on worker node\n' + wrapperscript)
    scriptName = "rootwrapper_generated_%s.py" % randomString()
    runScript = FileBuffer(scriptName, wrapperscript, executable=1)

    inputsandbox = app._getParent().inputsandbox + [script]
    return runScript, inputsandbox, rootenv


def defaultScript():
    tmpdir = tempfile.mktemp()
    os.mkdir(tmpdir)
    fname = os.path.join(tmpdir, 'test.C')
    script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                                   'defaultRootScript.C')
    from Ganga.GPIDev.Lib.File import FileUtils
    script = FileUtils.loadScript(script_location, '')
    with open(fname, 'w') as f:
        f.write(script)
    return fname


def defaultPyRootScript():
    tmpdir = tempfile.mktemp()
    os.mkdir(tmpdir)
    fname = os.path.join(tmpdir, 'test.py')
    script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                                   'defaultPyRootScript.py')

    from Ganga.GPIDev.Lib.File import FileUtils
    default_script = FileUtils.loadScript(script_location, '')

    with open(fname, 'w') as f:
        f.write(default_script)
    return fname


def randomString():
    """Simple method to generate a random string"""
    from random import randint
    from string import ascii_uppercase, join

    def addToSample(sample, ascii_length):
        """Basically random.select but python2.2"""
        a = ascii_uppercase[randint(0, ascii_length - 1)]
        if not a in sample:
            sample.append(a)
        else:
            # passing by referance
            addToSample(sample, ascii_length)

    ascii_length = len(ascii_uppercase)
    sample = []
    for _ in range(6):
        addToSample(sample, ascii_length)
    assert(len(sample) == 6)

    # seed is set to clock during import
    return join([str(a) for a in sample], '')





#!/usr/bin/env python
from __future__ import print_function
'''Script to run root with cint or python.'''
def downloadAndUnTar(fileName, url):
    '''Downloads and untars a file with tar xfz'''
    from shutil import copyfileobj
    from urllib2 import urlopen

    urlFileIn  = urlopen(url)
    urlFileOut = file(fileName,'w')
    copyfileobj(urlFileIn, urlFileOut)
    urlFileOut.close

    status = 0
    cmd = 'tar xzf %s' % fileName

    from commands import getstatusoutput, getoutput

    #to check whether the folder name  is 'root' or 'ROOT'
    folderName = getoutput('tar --list --file ROOT*.tar.gz').split('/')[0]

    try:#do this in try as module is only unix
        #commmand approach removes ugly tar error
        (status,output) = getstatusoutput(cmd)
    except ImportError:
        import os
        status = os.system(cmd)

    return status, folderName

def setEnvironment(key, value, update=False):
    '''Sets an environment variable. If update=True, it preends it to
       the current value with os.pathsep as the seperator.'''
    from os import environ, pathsep
    if update and key in environ:
        value += (pathsep + environ[key])#prepend
    environ[key] = value

def findPythonVersion(arch,rootsys):
    '''Digs around in rootsys for config files and then greps
       the version of python used'''
    import os

    def lookInFile(config):
        '''Looks in the specified file for the build config
           and picks out the python version'''
        version = None
        if os.path.exists(config):
            configFile = file(config)
            for line in configFile:#loop through the file looking for #define
                if line.startswith('#define R__CONFIGUREOPTION'):
                    for arg in line.split(' '):#look at value of #define
                        if arg.startswith('PYTHONDIR'):
                            arglist = arg.split('/')
                            if arglist[-1] == arch:
                                version = arglist[-2]
        return version

    def useRootConfig(rootsys):
        '''Use the new root-config features to find the python version'''
        version = None
        root_config = os.path.join(rootsys,'bin','root-config')
        if os.path.exists(root_config):
            import subprocess

            args = [root_config,'--python-version']

            run = subprocess.Popen(' '.join(args), shell = True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = [ e.splitlines() for e in run.communicate() ]
            code = run.returncode
            if code == 0 and out and not err and len(out) == 1:
                split = out[0].split('.')
                if len(out) != len(split):
                    version = '.'.join(split)
        return version

    version = None
    for f in ['config.h','RConfigure.h']:
        version = lookInFile(os.path.join(rootsys,'include',f))
        if version is not None:
            break
    if version is None:
        version = useRootConfig(rootsys)
    return version

def greaterThanVersion(version_string, version_tuple):
    '''Checks whether a version string is greater than a specific version'''
    result = False
    version_split = version_string.split('.')
    if len(version_split) == 3:
        try:
            major = int(version_split[0])
            minor = int(version_split[1])
            if major >= version_tuple[0] and minor > version_tuple[1]:
                result = True
        except:
            pass
    return result

def findArch(version):
    '''Method stub. In the future we might look at the
       environment to determin the arch we are running on.'''

    #SPI achitectures changed in Root > 5.16
    if greaterThanVersion(version, (5,16) ):
        return 'slc4_ia32_gcc34'
    return 'slc3_ia32_gcc323'

def findURL(version, arch):

    if greaterThanVersion(version, (5,16) ):
        fname = 'ROOT_%s__LCG_%s.tar.gz' % (version,arch)
    else:
        fname = 'root_%s__LCG_%s.tar.gz' % (version,arch)

    return fname



# Main
if __name__ == '__main__':

    from os import curdir, system, environ, pathsep, sep
    from os.path import join
    import sys

    commandline = ###COMMANDLINE###
    scriptPath = '###SCRIPTPATH###'
    usepython = ###USEPYTHON###

    version = '###ROOTVERSION###'
    arch = findArch(version)
    fname = findURL(version, arch)

    spiURL = 'http://service-spi.web.cern.ch/service-spi/external/distribution/'
    url = spiURL + fname

    print('Downloading ROOT version %s from %s.' % (version,url))
    (status, folderName) = downloadAndUnTar(fname,url)
    sys.stdout.flush()
    sys.stderr.flush()

    #see HowtoPyroot in the root docs
    import os
    pwd = os.environ['PWD']
    rootsys=join(pwd,folderName,version,arch,'root')
    setEnvironment('LD_LIBRARY_PATH',curdir,True)
    setEnvironment('LD_LIBRARY_PATH',join(rootsys,'lib'),True)
    setEnvironment('ROOTSYS',rootsys)
    setEnvironment('PATH',join(rootsys,'bin'),True)

    if usepython:

        pythonVersion = findPythonVersion(arch,rootsys)
        if not pythonVersion:
            print('Failed to find the correct version of python to use. Exiting', file=sys.stderr)
            sys.exit(-1)

        tarFileName = 'Python_%s__LCG_%s.tar.gz' % (pythonVersion, arch)
        url = spiURL + tarFileName

        print('Downloading Python version %s from %s.' % (pythonVersion,url))
        downloadAndUnTar(tarFileName,url)

        pythonDir = join('.','Python',pythonVersion,arch)
        pythonCmd = join(pythonDir,'bin','python')
        commandline = commandline % {'PYTHONCMD':pythonCmd}

        setEnvironment('LD_LIBRARY_PATH',join(pythonDir,'lib'),True)
        setEnvironment('PYTHONDIR',pythonDir)
        setEnvironment('PYTHONPATH',join(rootsys,'lib'),True)

    #exec the script
    print('Executing ',commandline)
    sys.stdout.flush()
    sys.stderr.flush()
    sys.exit(system(commandline)>>8)


